#include "forensic.h"

using namespace forensic;

/**
 * Creates a default cut definition.
 */
CutDefinition::CutDefinition()
{
  setExtension(".");
  maxLength = 0;
  setOutputDirectory("");
}

/**
 * Creates a duplicate of an existing cut definition.
 *
 * \param cut The cut definition to duplicate.
 */
CutDefinition::CutDefinition(const CutDefinition& cut)
{
  startRE = cut.startRE;
  endRE = cut.endRE;
  extension = cut.extension;
  maxLength = cut.maxLength;
  outputDirectory = cut.outputDirectory;
  comments = cut.comments;
}

/**
 * Destroys the cut definition.
 */
CutDefinition::~CutDefinition()
{
}

/**
 * Compares two cut definitions and returns true if they are equal.
 *
 * Two cut definitions are considered to be equal if
 * - the extensions match
 * - the start and end RE's are lexically the same respectively.
 * 
 * This is only a rough check and isn't considered to be 100% false-negative reliable.
 * (that is, the matching criteria is pretty strict and two RE's that are
 * functionally the same may be declared as being different)
 *
 * \param cut The cut definition to compare this against.
 * \returns True if the cut definitions match, false otherwise.
 */
bool CutDefinition::operator==(const CutDefinition& cut) const 
{
  // They're not the same if
  // - The extensions don't match
  // - the number of start and end RE's don't match
  if ((extension != cut.extension) ||
      (startRE.size() != cut.startRE.size()) ||
      (endRE.size() != cut.endRE.size()))
  {
    return false;
  }
  
  // - each RE isn't lexically the same
  // (note this is flawed, however its something)
  for (unsigned int i = 0; i < startRE.size(); i++)
  {
    bool found = false;
    for (unsigned int j = 0; (j < cut.startRE.size()) && (found == false); j++)
    {
      if (startRE[i] == cut.startRE[j])
      {
        found = true;
      }
    }
    
    if (found == false)
    {
      return false;
    }
  }
  
  for (unsigned int i = 0; i < endRE.size(); i++)
  {
    bool found = false;
    for (unsigned int j = 0; (j < cut.endRE.size()) && (found == false); j++)
    {
      if (endRE[i] == cut.endRE[j])
      {
        found = true;
      }
    }
    
    if (found == false)
    {
      return false;
    }
  }
  
  // Otherwise they're the same.
  return true;
}

/**
 * Makes this cut definition equal the supplied cut definition.
 *
 * \param cut The cut definition to copy.
 * \returns A reference to this.
 */
CutDefinition& CutDefinition::operator=(const CutDefinition& cut)
{
  if (this == &cut)
  {
    return *this;
  }
  
  startRE = cut.startRE;
  endRE = cut.endRE;
  extension = cut.extension;
  maxLength = cut.maxLength;
  outputDirectory = cut.outputDirectory;
  comments = cut.comments;
  
  return *this;
}

/**
 * Returns true if the cut definition has
 * - At least one starting RE
 * - A maximum length
 * - An extension
 * otherwise it returns false.
 *
 * \returns True if complete, false otherwise.
 */
bool CutDefinition::isValid() const
{
  bool allOk = true;
  
  if (startRE.size() == 0)
  {
    allOk = false;
  }
  
  if (extension.length() == 1)
  {
    allOk = false;
  }
  
  if (maxLength == 0)
  {
    allOk = false;
  }
  
  if (allOk == false)
  {
    cerr << "Incomplete CutDefinition:" << endl
         << "   Filename Extension     = " << extension.substr(1) << endl
         << "   Number of Starting REs = " << startRE.size() << endl
         << "   Maximum Length         = " << maxLength << endl
         << "All of the above should have valid values." << endl;
  }
  
  return allOk;
}

/**
 * Adds a start RE to this cut definition.
 *
 * The start RE should be bounded by forward slashes '/'.  All
 * characters outside the bound will be ignored (this includes 'i'
 * for case insensitive matches).
 *
 * \param re The regular expression.
 * \returns True if the regular expression passes simple sanity checking, 
 *          False otherwise.
 */
bool CutDefinition::addStartRE(const string& re)
{
  string filteredRE = filterRE(re);
  
  if (checkRE(filteredRE) == false)
  {
    return false;
  }
  
  for (unsigned int i = 0; i < startRE.size(); i++)
  {
    if (startRE[i] == filteredRE)
    {
      // It's already present.
      return true;
    }
  }
  
  startRE.push_back(filteredRE);
  
  return true;
}

/**
 * Returns a pointer to a vector containing all of the start RE's.
 *
 * Do not attempt to delete the pointer.
 *
 * \returns A vector pointer.
 */
const vector<string>* CutDefinition::getStartREs() const
{
  return &startRE;
}

/**
 * Adds a end RE to this cut definition.
 *
 * The end RE should be bounded by forward slashes '/'.  All
 * characters outside the bound will be ignored (this includes 'i'
 * for case insensitive matches).
 *
 * \param re The regular expression.
 * \returns True if the regular expression passes simple sanity checking, 
 *          False otherwise.
 */
bool CutDefinition::addEndRE(const string& re)
{
  string filteredRE = filterRE(re);
  
  if (checkRE(filteredRE) == false)
  {
    return false;
  }
  
  for (unsigned int i = 0; i < endRE.size(); i++)
  {
    if (endRE[i] == filteredRE)
    {
      // It's already present.
      return true;
    }
  }
  
  endRE.push_back(filteredRE);
  
  return true;
}

/**
 * Returns a pointer to a vector containing all of the end RE's.
 *
 * Do not attempt to delete the pointer.
 *
 * \returns A vector pointer.
 */
const vector<string>* CutDefinition::getEndREs() const
{
  return &endRE;
}

/**
 * Sets the extension for the cut definition.
 *
 * The extension will automaticall be prepended by a period '.' if
 * one is not already present.
 *
 * \param ext The extension
 */
void CutDefinition::setExtension(const string& ext)
{
  if (ext[0] != '.')
  {
    extension = "." + ext;
  } else
  {
    extension = ext;
  }
}

/**
 * Returns the extension for the cut definition.
 *
 * The extension will always have a leading period.
 *
 * \returns The extension.
 */
const string* CutDefinition::getExtension() const
{
  return &extension;
}

/**
 * Sets the maximum length of the cut definition.
 *
 * The length should be greater than zero.
 *
 * \param length The new maximum length
 */
void CutDefinition::setMaxLength(unsigned long length)
{
  if (length != 0)
  {
    maxLength = length;
  }
}

/**
 * Returns the maximum length for the cut definition.
 *
 * \returns The length;
 */
unsigned long CutDefinition::getMaxLength() const
{
  return maxLength;
}

/**
 * Sets the directory where the cuts will be placed.
 *
 * If the output directory is a string of zero length the present
 * working directory is assumed.
 *
 * \param outdir The directory.
 */
void CutDefinition::setOutputDirectory(const string& outdir)
{
  if (outdir.length() == 0)
  {
    outputDirectory = "./";
  } else if (outdir[outdir.length() - 1] != '/')
  {
    outputDirectory = outdir + "/";
  } else
  {
    outputDirectory = outdir;
  }
}

/**
 * Returns the output directory.
 *
 * The output directory is guaranteed to have a trailing slash.
 *
 * \returns A pointer to a string containing the output directory.
 */
const string* CutDefinition::getOutputDirectory() const
{
  return &outputDirectory;
}

/**
 * Adds a comment to this cut definition.
 *
 * The comment is any free flowing text.
 *
 * \param comment The comment.
 */
void CutDefinition::addComment(const string& comment)
{
  comments.push_back(comment);
}

/**
 * Returns a pointer to a vector of strings containing comments for this 
 * cut definition.
 *
 * The comments are any free flowing text attached to this cut definition.
 *
 * \returns The vector pointer.
 */
const vector<string>* CutDefinition::getComments() const
{
  return &comments;
}

/**
 * This method filters the provided RE to help ensure it conforms to our
 * standards.
 *
 * A new string is created and returned which contains the filtered RE.
 *
 * \param re The regular expression.
 * \returns A filtered regular expression.
 */
string CutDefinition::filterRE(const string& re) const
{
  char delimiter = re[0];
  unsigned int startPos = 0;
  unsigned int endPos = 1;
  
  while ((endPos < re.length()) && (re[endPos] != delimiter))
  {
    if (re[endPos] == '\\')
    {
      endPos++;
    }
    endPos++;
  }
  
  if (re[endPos] != delimiter)
  {
    cerr << "No closing delimiter (" << delimiter << ") found for " << re << endl;
    return NULL;
  }
  
  while (re[startPos + 1] == '^')
  {
    startPos++;
  }
  
  return re.substr(startPos + 1, endPos - 1);
}

/**
 * This method checks the provided RE to ensure that it conforms to RE standards.
 *
 * \param re The RE to be checked
 * \returns True if the RE is good, false otherwise.
 */
bool CutDefinition::checkRE(const string& re) const
{
  const char* errPtr;
  int errOffset;
  pcre* temp = pcre_compile(re.c_str(), 0, &errPtr, &errOffset, NULL);

  if (temp == NULL)
  {
    cerr << re << endl;
    cerr << errPtr << " at offset " << errOffset << endl;
    return false;
  }
  
  pcre_free(temp);
  
  return true;
}

