#include "forensic.h"

using namespace forensic;

/**
 * Returns the version of the forensic library.
 *
 * \returns The version as a C style string.
 */
const char* forensic::getVersion()
{
  return "0.5";
}

/**
 * Returns the date the Exgrep library was installed.
 *
 * \returns The date as a C style string.
 */
const char* forensic::getCompiledDate()
{
  return _COMPILED_ON;
}

/**
 * Creates an Exgrep object with the supplied configuration file.
 *
 * \param c The configuration file.
 */
Exgrep::Exgrep(const Config& c)
{
  config = &c;
  buffer = NULL;
  maxLength = 0;
  bufferLength = 0;
  bufferFillLength = 0;
  bufferStartPos = 0;
  inputStreamFileDescriptor = -1;
  cutDefinitionPos = 0;
  startOffset = 0;
  cutLength = 0;
  
  const vector<CutDefinition>* cuts = c.getActiveCutDefinitions();
  
  if (config->getVerbose() != 0)
  {
    cout << "Exgrep configured to match on the following file types:" << endl;
    
    map<string, bool> uniqueFilter;
    
    for (unsigned int i = 0; i < cuts->size(); i++)
    {
      const string* extension = (*cuts)[i].getExtension();
      
      if (uniqueFilter.find(*extension) == uniqueFilter.end())
      {
        cout << *extension << " ";
        uniqueFilter[*extension] = true;
      }
    }
    cout << endl;
  }
  
  maxLength = config->getMaxLength();
  
  // Making sure that the maxLength is a multiple of the
  // sector size;
  if (config->getSectorSize() != 0)
  {
    maxLength += config->getSectorSize() - ((maxLength - 1) % config->getSectorSize()) - 1;
  }
  
  // Need at least 2 times the max length size.
  bufferLength = maxLength * 2;
  buffer = new char[bufferLength];
  
  if (cuts->size() == 0)
  {
    cerr << "No CutDefinition's supplied!" << endl;
    exit(-1);
  }
  
  bool reFail = false;
  const char* error;
  int errorOffset;
  
  for (unsigned int i = 0; i < cuts->size(); i++)
  {
    string re = "";
    const vector <string>* startREs = (*cuts)[i].getStartREs();
    for (unsigned int j = 0; j < startREs->size(); j++)
    {
      if (j > 0)
      {
        re += "|";
      }
      
      re += (*startREs)[j];
    }
    
    pcre* reCompiled = pcre_compile(re.c_str(), 0, &error, &errorOffset, NULL);
    if (reCompiled == NULL)
    {
      reFail = true;
      cerr << re << endl;
      cerr << error << " at offset " << errorOffset << endl;
      continue;
    }
    
    cutStartPCRE.push_back(reCompiled);
    cutStartPCRE_Extra.push_back(pcre_study(reCompiled, 0, &error));
    
    if (error != NULL)
    {
      reFail = true;
      cerr << "Studying " << re << endl;
      cerr << error << endl;
      continue;
    }
    
    const vector <string>* endREs = (*cuts)[i].getEndREs();
    re = "";
    for (unsigned int j = 0; j < endREs->size(); j++)
    {
      if (j > 0)
      {
        re += "|";
      }
      re += ".*" + (*endREs)[j];
    }
    
    if (re.length() == 0)
    {
      cutEndPCRE.push_back(NULL);
      cutEndPCRE_Extra.push_back(NULL);
    } else
    {
      reCompiled = pcre_compile(re.c_str(), 0, &error, &errorOffset, NULL);
      if (reCompiled == NULL)
      {
        reFail = true;
        cerr << re << endl;
        cerr << error << " at offset " << errorOffset << endl;
        continue;
      }
      
      cutEndPCRE.push_back(reCompiled);
      cutEndPCRE_Extra.push_back(pcre_study(reCompiled, 0, &error));
      
      if (error != NULL)
      {
        reFail = true;
        cerr << "Studying " << re << endl;
        cerr << error << endl;
        continue;
      }
    }
  }
  
  if (reFail == true)
  {
    exit(-1);
  }
}

/**
 * Destroys the Exgrep object.
 */
Exgrep::~Exgrep()
{
  delete [] buffer;
  
  for (unsigned int i = 0; i < cutStartPCRE.size(); i++)
  {
    pcre_free(cutStartPCRE[i]);
  }
  
  for (unsigned int i = 0; i < cutStartPCRE_Extra.size(); i++)
  {
    pcre_free(cutStartPCRE_Extra[i]);
  }
  
  for (unsigned int i = 0; i < cutEndPCRE.size(); i++)
  {
    pcre_free(cutEndPCRE[i]);
  }
  
  for (unsigned int i = 0; i < cutEndPCRE_Extra.size(); i++)
  {
    pcre_free(cutEndPCRE_Extra[i]);
  }
  
  // Closing off any file descriptors
  if (inputStreamFileDescriptor != -1)
  {
    close(inputStreamFileDescriptor);
  }
}

/**
 * Instructs exgrep to get ready to process the supplied file.
 *
 * If successful the next step would be to call hasNext()
 *
 * \param inputFilename The name of the input file.
 * \returns 0 if successful, otherwise a non-zero value.
 */
int Exgrep::process(const string& inputFilename)
{
  cout << "Processing " << inputFilename << endl;
  
  int fd = open(inputFilename.c_str(), O_LARGEFILE | O_RDONLY);
  
  if (fd == -1)
  {
    cerr << "Could not open file " << inputFilename << endl;
    return -1;
  }
  
  process(fd);
  
  return 0;
}

/**
 * Processes the file descriptor.
 *
 * \param fd The file descriptor.
 */
void Exgrep::process(int fd)
{
  // Closing off any other file descriptors
  if ((inputStreamFileDescriptor != -1) && 
      (inputStreamFileDescriptor != 0))
  {
    close(inputStreamFileDescriptor);
    inputStreamFileDescriptor = -1;
  }
  
  inputStreamFileDescriptor = fd;
  bufferStartPos = 0;
  cutDefinitionPos = 0;
  startOffset = 0;
  
  // Need to skip to the start point.
  if (bufferStartPos < config->getStartPos())
  {
    if (config->getVerbose())
    {
      cout << "Skipping " << config->getStartPos() << " bytes" << endl;
    }
    
    // Seeking into the stream if its a file.
    if (lseek(inputStreamFileDescriptor, config->getStartPos(), SEEK_SET) != -1)
    {
      bufferStartPos = config->getStartPos();
    } else
    {
      // Seek failed.  If its a pipe handle it differently...
      if (errno == ESPIPE)
      {
        while (bufferStartPos < config->getStartPos())
        {
          unsigned int amount = config->getStartPos();
          if (amount > bufferLength)
          {
            amount = bufferLength;
          }
          read(inputStreamFileDescriptor, buffer, amount);
          bufferStartPos += amount;
        }
      } else
      {
        perror("Start offset seek");
      }
    }
  }
  
  fillBuffer();
}

/**
 * Fills the end of the buffer with data.  It attempts to fill the buffer completely.
 *
 * \returns The number of bytes read.
 */
int Exgrep::fillBuffer()
{
  if (bufferFillLength < 0)
  {
    return 0;
  }
  
  int count = read(inputStreamFileDescriptor, buffer + bufferFillLength, getNextReadSize(bufferLength - bufferFillLength));
  if (count == -1)
  {
    cerr << strerror(errno) << endl;
    return 0;
  }
  
  bufferFillLength += count;
  
  // Zeroing out the rest of the buffer (to get rid of old data)
  memset(buffer + bufferFillLength, 0x00, bufferLength - bufferFillLength);
  return count;
}

/**
 * Shifts the buffer across by maxLength bytes.
 */
void Exgrep::shiftBuffer()
{
  bufferStartPos += maxLength;
  
  bufferFillLength -= maxLength;
  
  if (bufferFillLength > 0)
  {
    memmove(buffer, buffer + maxLength, bufferFillLength);
  }
}

/**
 * This method returns the number of bytes the next read process should read.
 *
 * \param max The maximum number of bytes the read process is asking for.
 * \returns The number of bytes the read process should read.
 */
int Exgrep::getNextReadSize(int max) const
{
  if (config->getEndPos() == 0)
  {
    return max;
  }
  
  int count = config->getEndPos() - (bufferStartPos + bufferFillLength);
  
  if (count < 0)
  {
    count = 0;
  }
  
  if (count > max)
  {
    count = max;
  }
  
  return count;
}

/**
 * Used to iterate through the data stream looking for cuts.
 *
 * If it returns true the next step would be to call the getNext* methods.
 *
 * If it returns false the end of the stream has been reached.
 *
 * \returns True if a cut has been found.  False if the data stream has ended.
 */
bool Exgrep::hasNext()
{
  if (inputStreamFileDescriptor == -1)
  {
    cerr << "ERROR: No file descriptor provided" << endl;
    return false;
  }
  
  int ovector[45];
  int ovecsize = 45;
  
  while (true)
  {
    for (;cutDefinitionPos < cutStartPCRE.size(); cutDefinitionPos++)
    {
      while (startOffset < maxLength)
      {
        int retval = pcre_exec(cutStartPCRE[cutDefinitionPos], cutStartPCRE_Extra[cutDefinitionPos], buffer, bufferFillLength, startOffset, 0, ovector, ovecsize);
        
        if (retval >= 0)
        {
          int cutStart = ovector[0];
          
          startOffset = cutStart + 1;
          
          if ((cutStart < (int)maxLength) &&
              ((config->getSectorSize() == 0) || ((cutStart % config->getSectorSize()) == 0)))
          {
            if (cutEndPCRE[cutDefinitionPos] == NULL)
            {
              cutLength = (*config->getActiveCutDefinitions())[cutDefinitionPos].getMaxLength();
              if (cutLength > (unsigned)(bufferFillLength - cutStart))
              {
                cutLength = bufferFillLength - cutStart;
              }
              return true;
            } else
            {
              cutLength = findEnd(buffer + cutStart, cutDefinitionPos);
              if (cutLength == 0)
              {
                cutLength = (*config->getActiveCutDefinitions())[cutDefinitionPos].getMaxLength();
                if (cutLength > (unsigned)(bufferFillLength - cutStart))
                {
                  cutLength = bufferFillLength - cutStart;
                }
                return true;
              } else
              {
                return true;
              }
            }
          }
        } else if (retval == PCRE_ERROR_NOMATCH)
        {
          // No matches in this buffer
          startOffset = maxLength;
        } else
        {
          cerr << "Error: ";
          switch(retval)
          {
          case PCRE_ERROR_NULL:
            cerr << "Either code or subject was passed as NULL, or ovector was NULL and ovecsize was not zero." << endl;
            break;
          case PCRE_ERROR_BADOPTION:
            cerr << "An unrecognized bit was set in the options argument." << endl;
            break;
          case PCRE_ERROR_BADMAGIC:
            cerr << "PCRE stores a 4-byte \"magic number\" at the start of the compiled code, to catch the case when it is passed a junk pointer.  This is the error it gives when the magic number isn't present." << endl;
            break;
          case PCRE_ERROR_UNKNOWN_NODE:
            cerr << "While running the pattern match, an unknown item was encountered in the compiled pattern. This error could be caused by a bug in PCRE or by overwriting of the compiled pattern." << endl;
            break;
          case PCRE_ERROR_NOMEMORY:
            cerr << "If a pattern contains back references, but the ovector that is passed to pcre_exec() is not big enough to remember the referenced substrings, PCRE gets a block of memory at the start of matching to use for this purpose.  If the call via pcre_malloc()  fails, this error is given. The memory is freed at the end of matching." << endl;
            break;
          default:
            cerr << "Unknown error code: " << retval << endl;
          }
          cerr << "Programming error.  Cannot continue" << endl;
          exit(0);
        }
      }
      startOffset = 0;
    }
    
    // Need a new buffer.
    cutDefinitionPos = 0;
    shiftBuffer();
    
    if (fillBuffer() == 0)
    {
      if (bufferFillLength <= 0)
      {
        return false;
      }
    }
  }
}

/**
 * Returns the CutDefinition object for the cut that was just found.
 *
 * \returns The CutDefinition object.
 */
const CutDefinition* Exgrep::getNextCutDefinition() const
{
  return &(*config->getActiveCutDefinitions())[cutDefinitionPos];
}

/**
 * Returns a pointer to a buffer of data which contains the cut.
 *
 * \returns A pointer.
 */
const char* Exgrep::getNextBuffer() const
{
  return buffer + startOffset - 1;
}

/**
 * Returns the size of the buffer supplied by getNextBuffer()
 *
 * \returns The size.
 */
unsigned int Exgrep::getNextBufferLength() const
{
  return cutLength;
}

/**
 * Returns the offset from the start of the data stream (in bytes) for
 * the last cut discovered.
 *
 * NOTE: Due to the way the search algorithm is implemented it is possible that
 * this method will return offsets less than the offsets previously discovered.
 * For example, the first cut discovered may be at 2000, the next at 2500 and the
 * following at 1500.
 *
 * \returns The offset.
 */
unsigned int Exgrep::getNextCutOffset() const
{
  return bufferStartPos + startOffset - 1;
}

/**
 * This method gets called whenever a cut definition start RE was found and a 
 * ending RE needs to be matched.
 *
 * \param cutBuffer The data buffer starting from cut byte 0.
 * \param cutID The ID of the cut that it matched.
 * \returns The length of the cut of interest, or 0 if no ending RE match was found.
 */
inline int Exgrep::findEnd(const char* cutBuffer, int cutID) const
{
  // Need RE to start with .*
  
  int ovector[45];
  int ovecsize = 45;
  
  int retval = pcre_exec(cutEndPCRE[cutID], cutEndPCRE_Extra[cutID], cutBuffer, (*config->getActiveCutDefinitions())[cutID].getMaxLength(), 0, 0, ovector, ovecsize);
  
  if (retval >= 0)
  {
    return ovector[1];
  } else if (retval == PCRE_ERROR_NOMATCH)
  {
    // No matches in this buffer
    return 0;
  } else
  {
    cerr << "Error: ";
    switch(retval)
    {
    case PCRE_ERROR_NULL:
      cerr << "Either code or subject was passed as NULL, or ovector was NULL and ovecsize was not zero." << endl;
      break;
    case PCRE_ERROR_BADOPTION:
      cerr << "An unrecognized bit was set in the options argument." << endl;
      break;
    case PCRE_ERROR_BADMAGIC:
      cerr << "PCRE stores a 4-byte \"magic number\" at the start of the compiled code, to catch the case when it is passed a junk pointer.  This is the error it gives when the magic number isn't present." << endl;
      break;
    case PCRE_ERROR_UNKNOWN_NODE:
      cerr << "While running the pattern match, an unknown item was encountered in the compiled pattern. This error could be caused by a bug in PCRE or by overwriting of the compiled pattern." << endl;
      break;
    case PCRE_ERROR_NOMEMORY:
      cerr << "If a pattern contains back references, but the ovector that is passed to pcre_exec() is not big enough to remember the referenced substrings, PCRE gets a block of memory at the start of matching to use for this purpose.  If the call via pcre_malloc()  fails, this error is given. The memory is freed at the end of matching." << endl;
      break;
    default:
      cerr << "Unknown error code: " << retval << endl;
    }
    cerr << "Programming error.  Cannot continue" << endl;
    exit(0);
  }
  
  return 0;
}
