%{
  #include <iostream>
  #include <string>
  #include <vector>
  
  #include "forensic.h"
  
  #define YYSTYPE string
  #include "Parser.tab.h"
  
  using namespace std;
  using namespace forensic;
  
  extern "C"
  {
    int yylex();
    int yyparse();
    static void yyunput(int, char*);
  }
  
  void count();
  
  Config* config;
  bool error = false;
  
  vector<string> startRE;
  vector<string> endRE;
  unsigned int maxLength = 0;
  string extension = "";
  string outputDirectory = "";
  vector<string> comments;
  
  int mode;
%}

W [[:alnum:]_]
B [[:blank:]]

%%

#[^\n]*    yylval=yytext; count();  return COMMENT;
{W}+       yylval=yytext; count();  return WORD;
{B}*={B}*  yylval=yytext; count();  return EQUALS;
{B}*\[{B}* yylval=yytext; count();  return OBRACE;
{B}*\]{B}* yylval=yytext; count();  return EBRACE;
{B}*\n     yylval=yytext; count();  return NEWLINE;
{B}+       yylval=yytext; count();  return SPACE;
.          yylval=yytext; count();  return UCHAR;

%%

void clearCutDefinition()
{
  startRE.clear();
  endRE.clear();
  maxLength = 0;
  extension = "";
  outputDirectory = "";
  comments.clear();
}

void flushCutDefinition()
{
  if ((startRE.size() > 0) ||
      (endRE.size() > 0) ||
      (maxLength > 0) ||
      (extension.length() > 0) ||
      (outputDirectory.length() > 0) ||
      (comments.size() > 0))
  {
    if (config->addCutDefinition(startRE, endRE, maxLength, extension, outputDirectory, comments) == false)
    {
      error = true;
    }
    
    clearCutDefinition();
  }
}

bool parse(Config* configPtr, const string& filename)
{
  mode = MODE_INVALID;
  
  config = configPtr;
  
  yyin = fopen(filename.c_str(), "r");
  
  if (yyin == NULL)
  {
    cerr << "Could not open config file! " << filename << endl;
    return false;
  }
  
  yyparse();
  fclose(yyin);
  
  flushCutDefinition();
  
  return error == false;
}

int lineCountPrev = 1;
int charCountPrev = 0;
int lineCount = 1;
int charCount = 0;

void count()
{
  lineCountPrev = lineCount;
  charCountPrev = charCount;
  
  for (unsigned int i = 0; i < yylval.length(); i++)
  {
    switch(yylval[i])
    {
    case '\n':
      lineCount++;
      charCount = 0;
      break;
    
    case '\t':
      charCount += 8 - (charCount % 8);
      break;
    
    default:
      charCount++;
    }
  }
}

void yyerror(const char *str)
{
  cerr << str << ": between line " << lineCountPrev << ",character " << charCountPrev 
       << " and line " << lineCount << ",character " << charCount << endl;
//  error = true;
}

void processHeader(const string& header)
{
  flushCutDefinition();
  
  if (strcasecmp(header.c_str(), "GLOBAL") == 0)
  {
    mode = MODE_GLOBAL;
  } else if (strcasecmp(header.c_str(), "GROUPS") == 0)
  {
    mode = MODE_GROUPS;
  } else
  {
    mode = MODE_CUT_DEF;
    extension = header;
  }
}

void processValue(const string& name, const string& value)
{
  unsigned long long number;
  
  if (value.length() == 0)
  {
    return;
  }
  
  switch (mode)
  {
  case MODE_GLOBAL:
    if (strcasecmp(name.c_str(), "START") == 0)
    {
      number = strtoull(value.c_str(), NULL, 0);
      if (number >= 0)
      {
        config->setStartPos(number);
      } else
      {
        cerr << "ERROR!  Line " << lineCountPrev << ". START value cannot be negative." << endl;
        error = true;
      }
    } else if (strcasecmp(name.c_str(), "END") == 0)
    {
      number = strtoull(value.c_str(), NULL, 0);
      if (number >= 0)
      {
        config->setEndPos(number);
      } else
      {
        cerr << "ERROR!  Line " << lineCountPrev << ". END value cannot be negative." << endl;
        error = true;
      }
    } else if (strcasecmp(name.c_str(), "ACTIVE") == 0)
    {
      if (value.length() > 0)
      {
        config->setActiveGroups(value);
      }
    } else if (strcasecmp(name.c_str(), "SECTOR") == 0)
    {
      number = atoi(value.c_str());
      if (number >= 0)
      {
        config->setSectorSize(number);
      } else
      {
        cerr << "ERROR!  Line " << lineCountPrev << ". SECTOR value cannot be negative." << endl;
        error = true;
      }
    } else
    {
      cerr << "ERROR!  Line " << lineCountPrev << ". Unknown name \"" << name << "\"" << endl;
      error = true;
    }
    break;
   
  case MODE_GROUPS:
    config->defineGroup(name, value);
    break;
  
  case MODE_CUT_DEF:
    if (strcasecmp(name.c_str(), "START") == 0)
    {
      startRE.push_back(value);
    } else if (strcasecmp(name.c_str(), "END") == 0)
    {
      endRE.push_back(value);
    } else if (strcasecmp(name.c_str(), "LENGTH") == 0)
    {
      number = strtoull(value.c_str(), NULL, 0);
      if (number > 0)
      {
        maxLength = number;
      } else
      {
        cerr << "ERROR!  Line " << lineCountPrev << ". LENGTH value must be greater than zero." << endl;
        error = true;
      }
    } else if (strcasecmp(name.c_str(), "DIRECTORY") == 0)
    {
      outputDirectory = value;
    } else if (strcasecmp(name.c_str(), "COMMENT") == 0)
    {
      comments.push_back(value);
    } else
    {
      cerr << "ERROR!  Line " << lineCountPrev << ". Unknown name \"" << name << "\"" << endl;
      error = true;
    }
    break;
  
  default:
    cerr << "ERROR!  Line " << lineCountPrev << ". Cannot specify " << name << " = " << value << " without specifying a header first." << endl;
    cerr << "Specify a [GLOBAL], [GROUPS], or [<extension>] header first." << endl;
    error = true;
  }
}
