#ifndef FORENSIC_H
#define FORENSIC_H

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>
#include <map>

#include <pcre.h>
// If you prefer RedHat you will need to comment out the line above and uncomment the line below.
//#include <pcre/pcre.h>

#include <fcntl.h> // Used for open
#include <sys/stat.h> // Used for open
#include <sys/types.h> // Used for open
#include <unistd.h> // Used for close
#include <errno.h> // Used for errno
#include <stdio.h> // Used for perror

using namespace std;

namespace forensic
{
  const char* getVersion();
  const char* getCompiledDate();

  /**
   * A CutDefinition is an object that contains all the relevant
   * information about a single cut.  In its simplest form a 
   * cut is defined by a starting regular expression (RE), an
   * extension and a maximum length.
   *
   * This cut definition is self contained and will clean up after
   * itself, therefore do not free any pointers that may be returned.
   */
  class CutDefinition
  {
  public:
    CutDefinition();
    CutDefinition(const CutDefinition& cut);
    ~CutDefinition();
    
    bool operator==(const CutDefinition& cut) const;
    CutDefinition& operator=(const CutDefinition& cut);
    
    bool isValid() const;
    
    bool addStartRE(const string& re);
    const vector<string>* getStartREs() const;
    
    bool addEndRE(const string& re);
    const vector<string>* getEndREs() const;
    
    void setExtension(const string& ext);
    const string* getExtension() const;
    
    void setMaxLength(unsigned long length);
    unsigned long getMaxLength() const;
    
    void setOutputDirectory(const string& outdir);
    const string* getOutputDirectory() const;
    
    void addComment(const string& comment);
    const vector<string>* getComments() const;
    
  private:
    string filterRE(const string& re) const;
    bool checkRE(const string& re) const;
    
    /** A vector of strings that contains all the starting RE's for this cut definition. */
    vector<string> startRE;
    /** A vector of strings that contains all the ending RE's for this cut definition. */
    vector<string> endRE;
    
    /** Holds the extension for this cut definition (always has a '.' as its first character). */
    string extension;
    /** Holds the maximum length this cut definition should match. */
    unsigned long maxLength;
    /** Holds the directory the cuts should be placed into. */
    string outputDirectory;
    
    /** Holds a vector of strings containing all the comments given for this cut definition. */
    vector<string> comments;
  };
  
  /**
   * This class is designed to hold all of the configuration information for an Exgrep
   * object.  It contains all of the cut definitions including those not active.
   */
  class Config
  {
  public:
    Config();
    Config(const Config& config);
    ~Config();
    
    Config& operator=(const Config& config);
    
    void clear();
    
    void setVerbose(int value);
    int getVerbose() const;
    
    void setStartPos(unsigned long long pos);
    unsigned long long getStartPos() const;
    
    void setEndPos(unsigned long long pos);
    unsigned long long getEndPos() const;
    
    void setActiveGroups(const string& active);
    const string* getActiveGroups() const;
    
    void setSectorSize(int size);
    int getSectorSize() const;
    
    bool defineGroup(const string& name, const string& value);
    
    bool addCutDefinition(const vector<string>& startRE, const vector<string>& endRE, unsigned int max, const string& extension, const string& outputDirectory, const vector<string>& comments);
    const vector<CutDefinition>* getAllCutDefinitions() const;
    const vector<CutDefinition>* getActiveCutDefinitions() const;
    
    unsigned int getMaxLength() const;
    
    bool isValid();
    
    void dumpConfig(string file);
    
  private:
    void addCutDefinition(const CutDefinition& re);
    
    bool activateGroups();
    
    void writeConfigCommentHeader(ostream& out) const;
    void writeConfigCommentGlobal(ostream& out) const;
    void writeConfigCommentGroup(ostream& out) const;
    void writeConfigCommentCutDefinition(ostream& out) const;
    
    void writeConfigTextGlobal(ostream& out) const;
    void writeConfigTextGroup(ostream& out);
    void writeConfigTextCutDefinition(ostream& out) const;
    
    string stringToUpper(string str) const;
    bool activateCutDefinition(const string& cut);
    
    /** The amount of verboseness. */
    int verbose;
    /** The starting offset. */
    unsigned long long startPos;
    /** The ending offset (when we stop processing the input stream). */
    unsigned long long endPos;
    /** The size of the sector. */
    int sectorSize;
    /** A string containing active group names. */
    string activeGroups;
    /** A map containing group => cut definition extension relationships. */
    map<string, string> groups;
    /** A vector containing a bunch of cut definitions. (Active or inactive)*/
    vector<CutDefinition> cuts;
    /** A vector containing a bunch of active cut definitions. */
    vector<CutDefinition> activeCuts;
    
    /** The overall maximum length of a cut definition */
    unsigned int maxLength;
  };
  
  /**
   * The Exgrep class is used to search through an input stream looking for cuts.
   *
   * A cut is simply a series of bytes which has been identified by a particular
   * cut definition.  Typically this is used to search for files from a raw disk
   * without using the filesystem.
   *
   * Exgrep expects to be passed a configuration file when it is first constructed.
   * It is then possible to feed it file references or file descriptors for it to
   * start searching.
   */
  class Exgrep
  {
  public:
    Exgrep(const Config& c);
    ~Exgrep();
    
    int process(const string& inputFilename);
    void process(int fd);
    
    bool hasNext();
    const CutDefinition* getNextCutDefinition() const;
    const char* getNextBuffer() const;
    unsigned int getNextBufferLength() const;
    unsigned int getNextCutOffset() const;
    
  private:
    int fillBuffer();
    void shiftBuffer();
    int getNextReadSize(int max) const;
    
    int findEnd(const char* cutBuffer, int cutID) const;
    
    /** The configuration object */
    const Config* config;
    /** The data buffer */
    char* buffer;
    /** The maximum length of the cut definition lengths.*/
    unsigned long maxLength;
    /** The length of the data buffer */
    unsigned long bufferLength;
    /** The number of bytes currently in the data buffer */
    long bufferFillLength;
    /** The the offset buffer[0] is with respect to the input stream */
    unsigned long long bufferStartPos;
    
    /** The file descriptor for the input stream */
    int inputStreamFileDescriptor;
    /** The state data for the search to pause/resume */
    unsigned int cutDefinitionPos;
    /** The state data for the search to pause/resume */
    unsigned int startOffset;
    /** The length of the last cut found */
    unsigned int cutLength;
    
    /** A vector containing the starting PCRE's */
    vector<pcre*> cutStartPCRE;
    /** A vector containing the extra information for the starting PCRE's */
    vector<pcre_extra*> cutStartPCRE_Extra;
    /** A vector containing the ending PCRE's */
    vector<pcre*> cutEndPCRE;
    /** A vector containing the extra information for the ending PCRE's */
    vector<pcre_extra*> cutEndPCRE_Extra;
  };
};

#endif
