#include "main.h"

/**
 * This is parses the command line parameters, creates the configuration
 * file and an Exgrep object and starts Exgrep.
 *
 * \param argc The size of the argument array.
 * \param argv The argument array.
 */
int main(int argc, char* argv[])
{
  int opt = 0;
  static struct option long_options[] =
  {
    {"file",       1, 0, 'f'},
    {"config",     1, 0, 'c'},
    {"zeroconfig", 0, 0, 'Z'},
    
    {"filestart",  1, 0, 'A'},
    {"fileend",    1, 0, 'B'},
    {"sectorsize", 1, 0, 'S'},
    
    {"active",     1, 0, 'a'},
    {"verbose",    1, 0, 'v'},
    
    {"start",      1, 0, 's'},
    {"end",        1, 0, 'e'},
    {"max",        1, 0, 'm'},
    {"extension",  1, 0, 'x'},
    {"outputdir",  1, 0, 'o'},
    {"comment",    1, 0, 'C'},
    
    {"dumpconfig", 1, 0, 'd'},
    {"license",    0, 0, 'l'},
    {"help",       0, 0, 'h'},
    
    // Deprecated options
    {"unmatched",  1, 0, 'u'},
    {"re-size",    1, 0, 'r'},
    {0, 0, 0, 0}
  };
  Config config;
  vector<string> inputFilename;
  vector<string> configFilename;
  bool zeroConfig = false;
  
  string fileStart = "";
  string fileEnd = "";
  string sectorSize = "";
  
  string activeString;
  int verbose = -1;
  
  bool commandLineCutDef = false;
  vector<string> startREs;
  vector<string> endREs;
  unsigned int max = 100000;
  string extension = "";
  string outputDirectory = "";
  vector<string> comments;
  
  string dumpConfig = "";
  bool help = false;
  
  while ((opt = getopt_long(argc, argv, "f:c:ZA:B:S:a:v:s:e:m:x:o:C:d:lhu:r:", long_options, NULL)) != -1)
  {
    switch (opt)
    {
    case 'f':
      inputFilename.push_back(optarg);
      break;
    case 'c':
      configFilename.push_back(optarg);
      break;
    case 'Z':
      zeroConfig = true;
      break;
    
    case 'A':
      fileStart = optarg;
      break;
    case 'B':
      fileEnd = optarg;
      break;
    case 'S':
      sectorSize = optarg;
      break;
    
    case 'a':
      activeString = optarg;
      break;
    case 'v':
      verbose = atoi(optarg);
      break;
    
    case 's':
      commandLineCutDef = true;
      startREs.push_back(optarg);
      break;
    case 'e':
      commandLineCutDef = true;
      endREs.push_back(optarg);
      break;
    case 'm':
      commandLineCutDef = true;
      max = atoi(optarg);
      break;
    case 'x':
      commandLineCutDef = true;
      extension = optarg;
      break;
    case 'o':
      commandLineCutDef = true;
      outputDirectory = optarg;
      break;
    case 'C':
      commandLineCutDef = true;
      comments.push_back(optarg);
      break;
    
    case 'd':
      dumpConfig = optarg;
      break;
    case 'l':
      showLicense();
      return 0;
    case 'h':
      help = true;
     break;
    
    // The deprecated options
    case 'u':
    case 'r':
      cerr << "exgrep doesn't support the -" << (char)opt << " option anymore.  Ignored." << endl;
      break;
    
    default:
      return -1;
    }
  }
  
  if (help)
  {
    showUsage(verbose);
    return 0;
  }
  
  if (verbose > -1)
  {
    config.setVerbose(verbose);
  }
  
  if (zeroConfig)
  {
    config.clear();
  }
  
  // Parsing all the config files in order.
  for (unsigned int i = 0; i < configFilename.size(); i++)
  {
    if (parse(&config, configFilename[i]) == false)
    {
      cerr << "Error parsing config file: " << configFilename[i] << endl;
      return -1;
    }
  }
  
  // Parsing the command line cut definition.
  if (commandLineCutDef)
  {
    if (config.addCutDefinition(startREs, endREs, max, extension, outputDirectory, comments) == false)
    {
      cerr << "Command line cut definition invalid." << endl;
      return -1;
    }
  }
  
  if (fileStart.length() > 0)
  {
    unsigned long long start = strtoull(fileStart.c_str(), NULL, 0);
    config.setStartPos(start);
  }
  
  if (fileEnd.length() > 0)
  {
    unsigned long long end = strtoull(fileEnd.c_str(), NULL, 0);
    config.setEndPos(end);
  }
  
  if (sectorSize.length() > 0)
  {
    int size = atoi(sectorSize.c_str());
    
    if (size >= 0)
    {
      config.setSectorSize(size);
    } else
    {
      cerr << "Sector size cannot be negative.  A sector size of 0 removes searching on sector boundaries." << endl;
      return -1;
    }
  }
  
  if (activeString.length() > 0)
  {
    config.setActiveGroups(activeString);
  }
  
  if (config.isValid() == false)
  {
    cerr << "Config not valid" << endl;
    return -1;
  }
  
  if (commandLineCutDef)
  {
    // Checking to see if the command line specified RE is active
    // Otherwise send out a warning.
    if (extension[0] != '.')
    {
      extension = "." + extension;
    }
    
    bool found = false;
    
    const vector<CutDefinition>* cuts = config.getActiveCutDefinitions();
    for (unsigned int i = 0; i < cuts->size(); i++)
    {
      if ((*(*cuts)[i].getExtension()) == extension)
      {
        found = true;
      }
    }
    
    if (found == false)
    {
      cerr << "WARNING!  The command line supplied cut definition is not part of the active cut set!" << endl;
    }
  }
  
  if (dumpConfig.length() > 0)
  {
    config.dumpConfig(dumpConfig);
    return 0;
  }
  
  // If we are here then we're almost ready to go.  Create an Exgrep
  // object and pass it a configuration.
  Exgrep exgrep(config);
  
  int retVal = 0;
  
  // Time to do the crunching.
  if (inputFilename.size() == 0)
  {
    cout << "Processing STDIN" << endl;
    
    exgrep.process(0);
    
    while (exgrep.hasNext())
    {
      const CutDefinition *cutDefinition = exgrep.getNextCutDefinition();
      const char* buffer = exgrep.getNextBuffer();
      unsigned int length = exgrep.getNextBufferLength();
      unsigned int offset = exgrep.getNextCutOffset();
      writeToFile(buffer, length, offset, *cutDefinition, config, "");
    }
  } else
  {
    string dirPrefix = "";
    
    for (unsigned int i = 0; (i < inputFilename.size()) && (retVal == 0); i++)
    {
      if (inputFilename.size() != 1)
      {
        char* basename = strrchr(inputFilename[i].c_str(), '/');
        
        if (basename == NULL)
        {
          dirPrefix = inputFilename[i] + "_exgrep/";
        } else
        {
          dirPrefix = basename + 1;
          dirPrefix = dirPrefix + "_exgrep/";
        }
      }
      
      retVal = exgrep.process(inputFilename[i]);
      
      while (exgrep.hasNext())
      {
        const CutDefinition *cutDefinition = exgrep.getNextCutDefinition();
        const char* buffer = exgrep.getNextBuffer();
        unsigned int length = exgrep.getNextBufferLength();
        unsigned int offset = exgrep.getNextCutOffset();
        writeToFile(buffer, length, offset, *cutDefinition, config, dirPrefix);
      }
    }
  }
  
  return retVal;
}

/**
 * Shows a usage.  Increasing 'verbose' increases the amount of information.
 *
 * \param verbose The level of verbosity.
 */
void showUsage(int verbose)
{
  cout << endl
       << "exgrep - a regular expression forensic extraction tool." << endl
       << "Version " << getVersion() << endl
       << "Compiled on " << getCompiledDate() << endl
       << "Usage: exgrep [options]" << endl
       << endl;
  
  if (verbose <= 0)
  {
    cout << "-f,--file STR       Specifies a file to be scanned. Can be used more than once." << endl
         << "-c,--config STR     Specifies a config file to use. Multiple config files will extend previous config files" << endl
         << "-Z,--zeroconfig     Do not start with a default configuration." << endl
         << endl
         << "-A,--filestart INT  Specifies how far into the input stream scanning should start." << endl
         << "-B,--fileend INT    Specifies how far into the input stream scanning should end." << endl
         << "-S,--sectorsize INT Specify a sector size to use in bytes." << endl
         << endl
         << "-a,--active STR     Activate a group of cuts." << endl
         << "-v,--verbose INT    Set the level of verbosity." << endl
         << endl
         << "Command line cut definition parameters" << endl
         << "-s,--start STR      An RE to mark the start of the required region." << endl
         << "-e,--end STR        An RE to mark the end of the required region." << endl
         << "-m,--max INT        Maximum number of bytes between start and end REs." << endl
         << "-x,--extension STR  Extension to append to items that match this cut definition." << endl
         << "-o,--outputdir STR  Specifies the directory to place the cuts into." << endl
         << "-C,--comment STR    Attaches a comment to the command line cut." << endl
         << endl
         << "Others" << endl
         << "-d,--dumpconfig STR Creates a file containing the working config." << endl
         << "-l,--license        Prints out the license terms for this product." << endl
         << "-h,--help           Shows the help.  Increase verbosity for more help." << endl
         << endl;
  } else
  {
    cout << "-f,--file STR" << endl
         << "Specifies the input file to scan.  This parameter is optional and will default to "
         << "stdin. If specified multiple times exgrep will scan each file in turn, skipping "
         << "files that it cannot read.  When multiple input files are specified the output of "
         << "each will be placed in its own directory. " << endl
         << "Example: " << endl
         << "exgrep -f file1 -f file2" << endl
         << "exgrep will scan file1 first sending its output into the file1_exgrep directory. "
         << "It will then scan file2 sending its output into file2_exgrep." << endl
         << endl
         
         << "-c,--config STR" << endl
         << "Specifies a config file to use in addition to the internal default config (unless "
         << "-Z is specified)  It is valid to specify multiple configuration files.  Each "
         << "configuration file will overwrite or extend the previous configuration files." << endl
         << "Example: " << endl
         << "exgrep -c config1 -Z" << endl
         << "exgrep will ignore its default configuration and only use the configuration stored "
         << "in config1." << endl
         << "exgrep -c config1 -c config2" << endl
         << "exgrep will start with the default configuration.  It will then add config1 into "
         << "its configuration overwriting parts of the default config if they conflict.  It "
         << "will then read in config2 and overwrite its internal configuration if any conficts "
         << "are detected.  Therefore all of config2 will be present in the final configuration "
         << "while only the non-conflicting parts of config1 and the default config will be present."
         << endl
         
         << "-Z,--zeroconfig" << endl
         << "Informs exgrep that it is not to use its default configuration.  exgrep must be given "
         << "new configuration information via the -Z or -s options." << endl
         << "Example: " << endl
         << "exgrep -s /RE/ -m 100 -x fun -Z" << endl
         << "exgrep will discard its internal configuration and only search for this single cut "
         << "definition." << endl
         << endl
         
         << "-A,--filestart INT" << endl
         << "This value specifies how far into the input stream scanning should start.  The first "
         << "INT bytes will be ignored as if they didn't exist.  The default option is to start at "
         << "the first byte (0). If sector size scanning is in place this will cause the sector "
         << "size to START at the point specified." << endl
         << "Example: " << endl
         << "exgrep -A 100" << endl
         << "exgrep will discard the first 100 bytes and start scanning.  The 101th byte will be "
         << "considered byte 0, therefore the next sector boundary will actually occur at byte "
         << "612 (if the default sector size is 512 bytes)" << endl
         << endl
         
         << "-B,--fileend INT" << endl
         << "Specifies how far into the input stream scanning should end.  This value allows only "
         << "a portion of the input file to be searched.  This value must also be greater than the "
         << "filestart value, or it must be zero.  A zero indicates that the file will be scanned "
         << "until the end-of-file is reached.  The default value is zero." << endl
         << "Example: " << endl
         << "exgrep -B 500" << endl
         << "exgrep will only scan the first 500 bytes of the input stream (stdin) and then stop." << endl
         << endl
         
         << "-S,--sectorsize INT" << endl
         << "Specifies a sector size to use.  This means that the cuts must match on sector "
         << "boundaries.  This increases the likelihood that cuts found are correctly identified "
         << "however it may result in some cuts being missed.  The default value is 0 which means that "
         << "no sector size is used and a thorough search is performed." << endl
         << "Example: " << endl
         << "exgrep -S 512" << endl
         << "Tells exgrep to search on 512 byte boundaries." << endl
         << endl
         
         << "-a,--active STR" << endl
         << "Activate a group of cuts (comma separated).  By default the absence of this "
         << "parameter will cause all cuts to activate, however using this parameter will only "
         << "activate the cuts specified.  This parameter should only be used once." << endl
         << "Example: " << endl
         << "exgrep -a VIDEO,pdf" << endl
         << "exgrep will search for all cuts that are a part of the VIDEO group as well as "
         << "pdf files" << endl
         << endl
         
         << "-v,--verbose INT" << endl
         << "Sets the level of verbosity.  By default a value of 0 is assumed which is the "
         << "minimal amount of verboseness allowed.  A verbose level of 1 will increase the "
         << "amount of output sent to the screen while a verbose level of 2 will include all "
         << "the output of level 1 plus a bit extra.  Valid verbose values are 0, 1, 2 and 3" << endl
         << "Example: " << endl
         << "exgrep -v 1" << endl
         << "exgrep will increase its level of verbosity to level 1." << endl
         << endl
         
         << "-s,--start STR" << endl
         << "Specifies the starting RE of a command line provided cut definition.  When specifying a "
         << "command line cut definition it is also important to specify an extension (-x).  It is also "
         << "highly recommended that a maximum length is provided." << endl
         << "Example: " << endl
         << "exgrep -s /RE/ -x fun" << endl
         << "exgrep will add this rule for .fun filetypes into its default config and search stdin " << endl
         << endl
         
         << "-e,--end STR" << endl
         << "Specifies the ending pattern for the command line provided cut definition.  Must accompany both "
         << "-s and -x parameters.  The cuts will have a maximum length (by default 100000). If "
         << "the ending RE is found multiple times within this maximum length then the last match "
         << "is the one taken (ie the cuts are greedy).  If the ending RE is not found then the cut "
         << "will be the starting RE up until the maximum length." << endl
         << "Example: " << endl
         << "exgrep -s /RE/ -e /MODEL/ -x fun" << endl
         << "exgrep will search for a cut which starts with RE and optionally ends with MODEL up "
         << "to a maximum default length of 100000 bytes" << endl
         << endl
         
         << "-m,--max INT" << endl
         << "Specifies the maximum number of bytes between start and end REs.  This value " 
         << "is inclusive of the total number of bytes the RE's will match themselves.  If the "
         << "end RE is not found in the stream within this bound, or if the end RE is not "
         << "specified then the cut will be the start RE up until the maximum length" << endl
         << "Example: " << endl
         << "exgrep -s /RE/ -e /MODEL/ -m 10" << endl
         << "exgrep will search the input streams for something that starts with RE, ends with "
         << "MODEL and has at most 10 characters.  If the word REMODEL is present in the stream "
         << "the cut produced will be \"REMODEL\". If RE123MODEL is found the cut will be "
         << "\"RE123MODEL\".  However if RE1234MODEL is found then the cut will be RE1234MODE as "
         << "the cut is limited to its maximum size.  If RE1234567890 is found the cut will again "
         << "be RE12345678." << endl
         << endl
         
         << "-x,--extension STR" << endl
         << "Extension to appends to items that match this cut definition.  This parameter is mandatory if "
         << "a command line cut definition is specified.  The extension should not have a leading period "
         << "and ideally should be lower case (upper case is generally reserved for groups)." << endl
         << "Example: " << endl
         << "exgrep -s /RE/ -x fun" << endl
         << "exgrep will write any cuts found with this starting RE as a .fun file." << endl
         << endl
         
         << "-o,--outputdir STR" << endl
         << "Specifies the directory where all cuts matching this command line RE will be "
         << "placed." << endl
         << "Example: " << endl
         << "exgrep -s /RE/ -x fun -p dir1" << endl
         << "Any cuts matched by this definition will be written to the directory 'dir1'" << endl
         << endl
         
         << "-C,--comment STR" << endl
         << "Specifies a comment string to associate with the command line cut definition. "
         << "This value may be specified multiple times in order to allow for a larger "
         << "comment.  Note that comments may be interpreted by the shell before being "
         << "passed to exgrep so compensate accordingly." << endl
         << "Example: " << endl
         << "exgrep -s /RE/ -x fun -C \"This is a comment\"" << endl
         << "Will attach the comment 'This is a comment' to the command line cut definition." << endl
         << endl
         
         << "-d,--dumpconfig STR" << endl
         << "Creates a file containing the working config.  This is ideal to create new "
         << "configuration files based on the amalgamation of several config files. Also this "
         << "provides a level of debugging in order to see how exgrep has interpreted your "
         << "command line arguements." << endl
         << "Example: " << endl
         << "exgrep -d config1" << endl
         << "exgrep will dump its default config file to a file called config1" << endl
         << "exgrep -c config1 -c config2 -c config3 -s /RE/ -x fun -d configALL" << endl
         << "exgrep will dump the final configuration (after parsing all the other configuration "
         << "files) into the file configALL" << endl
         << endl
         
         << "-l,--license" << endl
         << "Prints out the license terms for exgrep." << endl
         << endl
         
         << "-h,--help" << endl
         << "Shows this help.  Turn off verbosity to see a simple help page" << endl
         << endl
         
         << "CONFIGURATION FILE PARAMETERS" << endl
         << "Refer to a exgrep generated configuration file to see what parameters are "
         << "available (use -d to generate a configuration file)" << endl
         << endl
         
         << "NOTE.  Command line parameters will always override config parameters.  Also "
         << "configuration file paramters will always override default exgrep configuration "
         << "parameters." << endl
         << endl;
  }
}

/**
 * Displays the license.
 */
void showLicense()
{
  cout << "exgrep - a regular expression forensic extraction tool" << endl
       << "Type exgrep -h for help." << endl
       << endl
       << "Copyright (C) 2003, Michael Cohen, Eddie Cornejo" << endl
       << endl
       << "This program is free software; you can redistribute it and/or"
          " modify it under the terms of the GNU General Public License"
          " as published by the Free Software Foundation; either version"
          " 2 of the License, or (at your option) any later version." << endl
       << endl
       << "This program is distributed in the hope that it will be useful,"
          " but WITHOUT ANY WARRANTY; without even the implied warranty of"
          " MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the"
          " GNU General Public License for more details." << endl
       << endl
       << "You should have received a copy of the GNU General Public"
          " License along with this program; if not, write to the Free"
          " Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,"
          " MA 02111-1307 USA" << endl;
}

/**
 * Writes the cut to a file.
 *
 * If the filename already exists it appends a 1up number to the end of
 * it until the name is unique.
 *
 * \param data The start of the cut.
 * \param length The length of the cut.
 * \param offset The offset in bytes from the start of the data stream.
 * \param cut A reference to the cut definition that matched it.
 * \param config The configuration settings for Exgrep.
 * \param directoryPrefix The directory to prepend to all cuts. Must end with a trailing slash.
 */
void writeToFile(const char* data, unsigned long length, unsigned long long offset, const CutDefinition& cut, const Config& config, const string& directoryPrefix)
{
  string filename = "";
  
  if ((*cut.getOutputDirectory())[0] == '/')
  {
    filename = "/" + directoryPrefix + cut.getOutputDirectory()->substr(1);
  } else
  {
    filename = directoryPrefix + (*cut.getOutputDirectory());
  }
  
  int dirfd = open(filename.c_str(), O_DIRECTORY);
  close(dirfd);
  
  if (dirfd == -1)
  {
    if (config.getVerbose())
    {
      cerr << "Directory \"" << filename << "\" does not exist." << endl;
      cerr << "Attempting to create..." << endl;
    }
    
    if (makeDirectory(filename) == false)
    {
      if (config.getVerbose())
      {
        cerr << "Could not create directory!" << endl;
      }
      cerr << "Aborting write to file as the directory \"" << filename << "\" could not be created" << endl;
      return;
    }
  }
  
  char loc[17];
  snprintf(loc, 17, "%016llu", offset);
  filename += loc;
  cout << cut.getExtension()->substr(1) << " found at " << loc << endl;
  
  string basefilename = filename;
  filename = basefilename + (*cut.getExtension());
  
  int fd = open(filename.c_str(), O_RDONLY);
  
  int count = 1;
  
  while (fd != -1)
  {
    // File exists.  Must find a unique filename
    close(fd);
    filename = basefilename + "_";
    snprintf(loc, 17, "%d", count++);
    filename += loc;
    filename += (*cut.getExtension());
    fd = open(filename.c_str(), O_RDONLY);
  }
  
  close(fd);
  fd = open (filename.c_str(), O_CREAT|O_WRONLY, 0666);
  if (fd == -1)
  {
    cerr << "WARNING: Could not create file \"" << filename << "\"" << endl;
    return;
  }
  
  write(fd, data, length);
  close(fd);
}

/**
 * Makes a directory.
 *
 * This method will also attempt to create and parent directories that don't
 * currently exist.
 *
 * \param directory The directory name to be created.
 * \returns True if the directory was made.  False otherwise.
 */
bool makeDirectory(string directory)
{
  int pos = 0;
  int lastPos = 0;
  
  if (directory[directory.length() - 1] == '/')
  {
    directory = directory.substr(0, directory.length() - 1);
  }
  
  string dir = "";
  
  do
  {
    pos = directory.find_first_of("/", lastPos);
    
    dir += directory.substr(lastPos, pos - lastPos) + "/";
    
    int dirfd = open(dir.c_str(), O_DIRECTORY);
    close(dirfd);
    
    if (dirfd == -1)
    {
      if (mkdir(dir.c_str(), 0777))
      {
        return false;
      }
    }
    lastPos = pos + 1;
    
  } while (pos != -1);
  
  return true;
}
