#include "index.h"
#include <stdio.h>
#include "except.h"
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#define BUFFER 1000

static unsigned char Malloc[]="Malloc error";
// The entire index is managed within a single malloced chunk. This
// makes it easier to serialise to disk and back again. We implement a
// psuedo heap management facility which does not do free, only
// mallocs through the approriate constructors:
static void *heap=NULL;
static int lines=0;


// The root of the tree.
static heap_ptr root; 

//This represents the end of the current structures on the heap.
static heap_ptr end_of_heap_ptr=0;
static long int heap_size=0;

//The block size we grow our heap by (1mb by default).
static int heap_malloc_size=1024*1024;

/* Called to initialise our heap structures */
void init_heap() {
  heap=calloc(heap_malloc_size,1);
  if(!heap) RAISE(E_NOMEMORY,NULL,Malloc);
  
  heap_size=heap_malloc_size;
  end_of_heap_ptr=0;
};

/*Implements our own malloc giving a small chunk off our own heap - if
  our heap needs to grow, this will take care of it */
heap_ptr mymalloc(heap_ptr size) {
  end_of_heap_ptr+=size;
  //Do we fit in the currently allocated chunk?
  if(end_of_heap_ptr > heap_size) {
    heap_size+=heap_malloc_size;
    printf("Will try to allocate %lu (%u)\n",heap_size,lines);
    heap=realloc(heap,heap_size);
    if(!heap) RAISE(E_NOMEMORY,NULL,Malloc);
    memset(heap+heap_size-heap_malloc_size,0,heap_malloc_size);
  };

  return(end_of_heap_ptr);
};

/* Assigns a new node structure on our heap. Note that the real
   address in memory must be gotten by using absolute_node(heap_ptr)
*/
heap_ptr new_node() {
  return(mymalloc(sizeof(struct idx_node)));
};

heap_ptr new_offset_list() {
  return(mymalloc(sizeof(struct offset_list)));
};

/* Convert from internal heap offsets to memory pointers */
inline struct idx_node *absolute_node(heap_ptr relative) {
  return(struct idx_node *)(relative+heap);
};
inline struct offset_list *absolute_offlist(heap_ptr relative) {
  return(struct offset_list *)((heap_ptr)relative+heap);
};

/* Adds the word pointed to by c to the index at node */
void add_word(heap_ptr node_ref,unsigned char *c) {
  struct idx_node *node,*temp;
  heap_ptr temp_ptr;

  //Get memory pointers
  node=absolute_node(node_ref);
  
  //We reached the end of the word?
  if(!*(c)) {
    node->offsets=new_offset_list();
    return;
  };

  /* Walk over all peers to see if any of them have the char c */
  temp=node;
  while(1) {
    if(temp->c==tolower(*c)) {
      //Yep we found it. Does this node have a child?
      if(!*(c+1)) {
	temp->offsets=new_offset_list();
      } else if(temp->child) {
	add_word(temp->child,c+1);
      } else {
	struct idx_node *child;
	
	temp->child=new_node();
	child=absolute_node(temp->child);
	child->c=tolower(*(c+1));
	add_word(temp->child,c+1);
      };
      return;
    };

    if(!temp->peer) {
      //Thats the end of the line - we could not find it, create a new peer, and search for it again:
      struct idx_node *peer;
      
      temp->peer=new_node();
      peer=absolute_node(temp->peer);
      peer->c=tolower(*(c));
      //Try to add that letter again - this time we should find it.
      add_word(temp->peer,c);
      return;
    };

    temp=absolute_node(temp->peer);
  };
};

/* Asks if the word is in the index. Returns the heap_ptr for the block. 
   Note that root is a relative pointer to our heap.

   c will be advanced as much as is needed, so if a match occurs it
   will be pointing at the end of the string that matched. However if
   a match did not occur, it will point at as much of the words as
   matched.
*/
heap_ptr is_in_index(heap_ptr root_ptr,unsigned char **word) {
  struct idx_node *temp;
  struct idx_node *root;
  unsigned char *c= *word;
  heap_ptr temp_ptr,temp_ptr2;

  //Reset memory reference to absolute.
  root=absolute_node(root_ptr);

  //No root node - the word is not in the index.
  if(!root) return(0);

  //If this current node has an offset, we found at least some of the data:
  if(root->offsets && tolower(*c)==root->c) {
    (*word)++;
    return(root_ptr);
  };

  //Search for this char in all peers, if its there, we recursivly search for the next few chars in the children.
  for(temp_ptr=root_ptr;temp_ptr;temp_ptr=temp->peer) {
    temp=absolute_node(temp_ptr);
    if(temp->c==tolower(*c)) {
      //We got to the bottom of the tree - lets return this match.
      if(!temp->child) {
	(*word)++;
	return(temp_ptr);
      } else {
	//We found the node we are after
	(c)++;
	//A word terminates here, we return it.
	if(temp->offsets) {
	  (*word)++;
	  return(temp_ptr);
	};

	//See if the rest of the word matches
	temp_ptr2=is_in_index(temp->child,&c);
	if(temp_ptr2) {
	  *word=c;
	  return(temp_ptr2);
	  //No it doesnt match a full word:
	} else {
	  return(0);
	};
      };
    };
  };
  return(0);
};


/* Recurse over the whole tree and print the nodes: string is the
   start of the string that we are trying to build, pos is the
   posiiton within the string where we are currently at.
 */
void print_tree(heap_ptr root_ptr,char *string,char *pos) {
  struct idx_node *temp;
  heap_ptr temp_ptr;
  int j;
  struct offset_list *offs;

  if(!string) {
    string=(char *)malloc(BUFFER);
    pos=string;
  };
  if(!string) RAISE(E_NOMEMORY,NULL,Malloc);

  for(temp_ptr=root_ptr;temp_ptr;temp_ptr=temp->peer) {
    temp=absolute_node(temp_ptr);
    
    //Set this to get the word name that we are storing here.
    *(pos)=temp->c;

    //A word terminates here!!!
    if(temp->offsets) {
      *(pos+1)=0;
      //Count how many elements are in each list:
      j=0;
      offs=absolute_offlist(temp->offsets);
      while(1) {
	offs=absolute_offlist(offs->next);
	if(offs->offset==0 && offs->next==0) break;
	//	if(offs->offset) printf("Offset at %llu\n",offs->offset);
	j++;
      };
      if(j>0) 
	printf("%s -> %u  %u matches \n",string,temp->offsets,j);
    };
    
    //We have a child, we need to recurse here...
    if(temp->child) {
      print_tree(temp->child,string,pos+1);
    };
  };
};


/* Save the heap structure to disk */
void idx_save_to_file(unsigned char *name) {
  int fd;
  fd=creat(name,S_IRWXU);
  if(!fd) 
    RAISE(E_IOERROR,NULL,"Could not create file %s",name);

  if(write(fd,heap,end_of_heap_ptr)<end_of_heap_ptr) {
    close(fd);
    RAISE(E_IOERROR,NULL,"Could not write index to file %s",name);
  };

  close(fd);
};

/* Loads our heap from the file */
void idx_load_from_file(unsigned char *name) {
  int fd,result;
  fd=open(name,O_RDONLY);
  if(!fd) RAISE(E_IOERROR,NULL,"Cant open file %s",name);

  init_heap();
  while(1) {
    result=read(fd,heap+end_of_heap_ptr,heap_malloc_size);
    if(result<0) {
      close(fd);
      RAISE(E_IOERROR,NULL,"Cant read %s",name);
    };
    
    if(result==0) break;

    end_of_heap_ptr+=result;
    if(end_of_heap_ptr>=heap_size) {
      heap_size+=heap_malloc_size;
      heap=realloc(heap,heap_size);
      if(!heap) RAISE(E_NOMEMORY,NULL,Malloc);
    };
  };

  close(fd);
};

/* Follows the end of the list and adds a new offset to it */
void idx_add_offset_to_list(heap_ptr o,unsigned long long int offset) {
  struct offset_list *off=absolute_offlist(o);
  //Go to the end:
  while(off->next) {
    off=absolute_offlist(off->next);
  };

  off->next=new_offset_list();
  off=absolute_offlist(off->next);
  off->offset=offset;
};


/* This indexes the buffer:
   data,length represent a binary text string to index.
   base is the base offset of this string within the file.
 */
void idx_index_buffer(long long int base,unsigned char *data,int length) {
  unsigned char *c;
  int j;
  char *tmp=(char *)malloc(BUFFER);
  struct idx_node *node;
  heap_ptr i;

  for(j=0;j<length;j++) {
    //Point at the current char in the buffer
    c=data+j;
    //Search for the current word in the index
    i=root;
    //This continues searching for larger words when subwords are specified...
    for(i=is_in_index(i,&c);i;i=is_in_index(i,&c)) {
      strncpy(tmp,data+j,c-data-j);
      *(tmp+(c-data)-j)=0;

      //If there is a child in this node, we keep searching for longer matches
      node=absolute_node(i);
      printf("Added %s offset %llu to index\n",tmp,base+j);

      idx_add_offset_to_list(node->offsets,base+j);
      if(node->child) {
	i=node->child;
      } else {
	break;
      };
    };
  };
};


int main() {
  unsigned char *tmp=(char *)malloc(BUFFER);

  init_heap();
  root=new_node();
  /*
  while(fgets(tmp,BUFFER,stdin)) {
    int length=strlen(tmp)-1;
    
    lines++;
    if(tmp[length]=='\n') {
      tmp[length]=0;
    };
    if(strlen(tmp)>=3)
    add_word(root,tmp);
  };
  idx_save_to_file("/tmp/temp.idx");
  */
  idx_load_from_file("/tmp/temp.idx");
  tmp="indexing hey there, Hello, This is a short test of indexing hey ok?";
  idx_index_buffer(0,tmp,strlen(tmp));
  print_tree(root,NULL,0);

  exit(0);
  while(fgets(tmp,BUFFER,stdin)) {
    int length=strlen(tmp)-1;

    lines++;
    if(tmp[length]=='\n') {
      tmp[length]=0;
    };
    printf("%s is at %u\n",tmp,is_in_index(root,&tmp));
  };
  return(0);
};
