/*
** flag dbtool

   Author:         David Collett (daveco@users.sourceforge.net)
   Version: 0.1   
   Copyright (2004).

   This software traverses an image and produces SQL statements which
   can be forwarded to a database for offline analysis. This program
   forms part of pyflag, which can be found on:

   http://pyflag.sourceforge.net/

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
*/

#include "fs_tools.h"
#include "error.h"
#include "ntfs.h"
#include "fs_io.h"
#include "except.h"

/* Time skew of the system in seconds */
static int32_t sec_skew = 0;

char *tbl_name;
char *fstype;
int file_count = 0;
int inode_count = 0;
int inode_total = 0;

typedef struct file_run {
  unsigned long addr;
  INUM_T inum;
  int type;
  int id;
  struct file_run *next;
} RUN;

/* some prototypes */

void print_dent2(FILE *hFile, FS_DENT *fs_dent, int flags, FS_INFO *fs, 
  FS_DATA *fs_data);

static u_int8_t
print_inode(FS_INFO *fs, INUM_T inum, FS_INODE *fs_inode, int flags,
			        char *unused_context);
void print_sql_string(FILE *fh, const unsigned char *ptr, int length);

static u_int8_t
print_addr(FS_INFO *fs, DADDR_T addr, char *buf,
  int size, int flags, char *ptr);

void print_blocks(INUM_T inum, int type, int id, RUN *run);
void cleanup_deleted(char *name);
void cleanup(char *name);
void create_tables(char *name);
void drop_tables(char *name);

/*
 * print usage message
 */
void usage(char *myProg) {
	printf("usage: %s [-vV] [-t table_name] [-f fstype] [-z ZONE] [-s seconds] [-d (create|drop)] image\n", 
	  myProg);
	printf("\tIf [inode] is not given, the root directory is used\n");
	printf("\t-v: verbose output to stderr\n");
	printf("\t-i:  select IO Subsystems. Try help for a list of subsystems\n");
	printf("\t-t: output table name\n");
	printf("\t-V: Print version\n");
	printf("\t-z: Time zone of original machine (i.e. EST5EDT or GMT) (only useful with -l)\n");
	printf("\t-s seconds: Time skew of original machine (in seconds) (only useful with -l & -m)\n");
    	printf("\t-f fstype: Image file system type\n");
    	printf("\t-d (create|drop): Print create of drop table strings\n");
	printf("Supported file system types:\n");
	fs_print_types();

	exit(1);
}

FILE *logfp;

/* 
 * call back action function for dent_walk
 * calls print_dent2 to do actual printing, since it 
 * must be called once for each data stream in NTFS
 */
static u_int8_t
print_dent (FS_INFO *fs, FS_DENT *fs_dent, int flags, char *ptr) 
{

  // print progress message
  file_count++;
  if(file_count % 200 == 0)
    fprintf(stderr, "Loaded %d file entries\n", file_count);
        /* have to make a special case for NTFS */
	if (((fs->ftype & FSMASK) == NTFS_TYPE) && (fs_dent->fsi)) {

		FS_DATA *fs_data = fs_dent->fsi->attr;

		while ((fs_data) && (fs_data->flags & FS_DATA_INUSE)) {

			if (fs_data->type == NTFS_ATYPE_DATA) {
				mode_t mode = fs_dent->fsi->mode; 
				u_int8_t ent_type = fs_dent->ent_type;


				/* 
				 * A directory can have a Data stream, in which
				 * case it would be printed with modes of a
				 * directory, although it is really a file
				 * So, to avoid confusion we will set the modes
				 * to a file so it is printed that way.  The
				 * entry for the directory itself will still be
				 * printed as a directory
				 */

				if ((fs_dent->fsi->mode & FS_INODE_FMT) == FS_INODE_DIR) {
					
					fs_dent->fsi->mode &= ~FS_INODE_FMT;	
					fs_dent->fsi->mode |= FS_INODE_REG;	
					fs_dent->ent_type = FS_DENT_REG;
				}
				
				print_dent2(stdout, fs_dent, flags, fs, fs_data);

				fs_dent->fsi->mode = mode; 
				fs_dent->ent_type = ent_type;
			}
			else if (fs_data->type == NTFS_ATYPE_IDXROOT) {
				if (!(ISDOT (fs_dent->name)))
					print_dent2(stdout, fs_dent, flags, fs, fs_data);
			}

			fs_data = fs_data->next;
		}

	}
	else {
		if (!(ISDOT (fs_dent->name)))
			print_dent2(stdout, fs_dent, flags, fs, NULL);
	}

	return WALK_CONT;
}

/*
 * Print one directory entry for file table
 */
void
print_dent2(FILE *hFile, FS_DENT *fs_dent, int flags, FS_INFO *fs, 
  FS_DATA *fs_data)
{
	char mode[4];
	char status[10] = "alloc";

	FS_INODE *fs_inode = fs_dent->fsi;
	
	/* type of file - based on dentry type */
	if ((fs_dent->ent_type & FS_DENT_MASK) < FS_DENT_MAX_STR)
		sprintf(mode, "%s/", fs_dent_str[fs_dent->ent_type & FS_DENT_MASK]);
	else {
		strcpy(mode, "-/");
	}

	/* type of file - based on inode type: we want letters though for
	 * regular files so we use the dent_str though */
	if (fs_inode) {
		int typ = (fs_inode->mode & FS_INODE_FMT)>>FS_INODE_SHIFT;
		if ((typ & FS_INODE_MASK) < FS_DENT_MAX_STR)
			strcat(mode, fs_dent_str[typ & FS_INODE_MASK]);
		else {
			strcat(mode, "-");
		}
	}
	else {
		strcat(mode, "-");
	}

	mode[3] = '\0';


	/* deleted */
	if (flags & FS_FLAG_UNALLOC) {
		if((fs_inode) && (fs_inode->flags & FS_FLAG_ALLOC)) {
			strcpy(status, "realloc");
		}
		else {
			strcpy(status, "deleted");
		}
	}

	if(strcmp(status, "alloc") != 0) {
		cleanup_deleted(fs_dent->name);
	}
	cleanup(fs_dent->name);

	if(fs_data)
	  /* print the data stream name if we the non-data NTFS stream */
	  if ( fs_data && (((fs_data->type == NTFS_ATYPE_DATA) && 
			    (strcmp(fs_data->name, "$Data") != 0)) ||
			   ((fs_data->type == NTFS_ATYPE_IDXROOT) && 
			    (strcmp(fs_data->name, "$I30") != 0))) )
	    //fprintf(hFile, "('%i','%i','%i','%s','%s','/%s%s:%s),'\n",
	    //fprintf(hFile, "INSERT INTO file_%s VALUES('%i','%i','%i','%s','%s','/%s','%s:%s');\n", tbl_name,
	    fprintf(hFile, "INSERT INTO file_%s VALUES('%i-%i-%i','%s','%s','/%s','%s:%s');\n", tbl_name,
		    fs_dent->inode, fs_data->type, fs_data->id, mode, status, fs_dent->path, 
		    fs_dent->name, fs_data->name);
	  else
	    //fprintf(hFile, "('%i','%i','%i','%s','%s','/%s%s),'\n",
	    //fprintf(hFile, "INSERT INTO file_%s VALUES('%i','%i','%i','%s','%s','/%s','%s');\n", tbl_name,
	    fprintf(hFile, "INSERT INTO file_%s VALUES('%i-%i-%i','%s','%s','/%s','%s');\n", tbl_name,
		    fs_dent->inode, fs_data->type, fs_data->id, mode, status, fs_dent->path, fs_dent->name);
	else {
	  //fprintf(hFile, "('%i','%i','%i','%s','%s','/%s%s'),\n",
	  //fprintf(hFile, "INSERT INTO file_%s VALUES('%i','%i','%i','%s','%s','/%s','%s');\n", tbl_name,
	  fprintf(hFile, "INSERT INTO file_%s VALUES('%i','%s','%s','/%s','%s');\n", tbl_name,
		  fs_dent->inode, mode, status, fs_dent->path, fs_dent->name);
	}
	return;
}

/*
 * function for cleaning filenames
 */
void cleanup_deleted(char *name) {
	int char_count = 0;
	int i=0;
	int len = strlen(name);

	for(i=0; i<len; i++) {
		if(name[i] < 20 || name[i] > 126) {
			char_count++;
		}
	}

	//if(char_count > len/2) {
	if(char_count > 0) {
		strncpy(name, "<UNKNOWN>", len);
	}
	return;
}

/*
 * sanitise data as req'd for mysql
 */
void cleanup(char *name) {
	int i;
	for(i=0; i<strlen(name); i++) {
		if(name[i] == 39) { // "'"
			name[i] = ' ';
		}
	}
}

/* 
 * call back action function for inode_walk
 */
static u_int8_t
print_inode(FS_INFO *fs, INUM_T inum, FS_INODE *fs_inode, int flags,
			        char *unused_context)
{
  time_t dtime = 0;
  char *link;

  // print progress message
  inode_count++;
  if(inode_count % 200 == 0)
    fprintf(stderr, "Loaded %d of %d Inodes\n", inode_count, inode_total);

	// is this a symlink?
	if(fs_inode->link != NULL) {
		link = strdup(fs_inode->link);
		cleanup_deleted(link);
		cleanup(link);
	} else
		link = strdup("");

	if((fs_inode->uid == 0) && (fs_inode->gid == 0) && 
			(fs_inode->mtime == 0) && (fs_inode->atime == 0) &&
	   		(fs_inode->ctime == 0) && (fs_inode->dtime == 0) &&
	   		(fs_inode->nlink == 0) && (fs_inode->size == 0)) 
		return WALK_CONT;

	// adjust times if necessary
	if (sec_skew != 0) {
		fs_inode->mtime -= sec_skew;
		fs_inode->atime -= sec_skew;
		fs_inode->ctime -= sec_skew;
	}

	// do we have dtime?
	if (fs->flags & FS_HAVE_DTIME) {
		if (sec_skew != 0) 
			fs_inode->dtime -= sec_skew;

		dtime = fs_inode->dtime;
    		//printf(",'%lu'", (ULONG) fs_inode->dtime);

		if (sec_skew != 0) 
			fs_inode->dtime += sec_skew;
	}
	//else
	  //printf(",'0'");

	//printf(",'%lo','%d','%s','%lu');\n",
	//   (ULONG) fs_inode->mode, (int) fs_inode->nlink, link,
	//   (ULONG) fs_inode->size);


	/* now print data blocks (for each data stream in NTFS) */
	if (((fs->ftype & FSMASK) == NTFS_TYPE) && (fs_inode)) {
		FS_DATA *fs_data;
		fs_data = fs_inode->attr;

	    	while ((fs_data) && (fs_data->flags & FS_DATA_INUSE)) {
	 	       	//if (fs_data->flags & FS_DATA_NONRES) {
			if(fs_data->type == NTFS_ATYPE_DATA) {
				RUN run;
				run.addr = -1;
				run.inum = inum;
				run.type = fs_data->type;
				run.id = fs_data->id;
				run.next = NULL;
				RUN *ptr = &run;

				//printf("INSERT INTO inode_%s VALUES('%lu','%d','%d','%c','%d','%d','%lu',"
				printf("INSERT INTO inode_%s VALUES('%lu-%d-%d','%c','%d','%d','%lu'," \
				       "'%lu','%lu',%lu,'%lo','%d','%s','%lu');\n",
				       tbl_name,
				       (ULONG) inum, fs_data->type, fs_data->id,
				       (flags & FS_FLAG_ALLOC) ? 'a' : 'f',
				       (int) fs_inode->uid, (int) fs_inode->gid,
				       (ULONG) fs_inode->mtime, (ULONG) fs_inode->atime,
				       (ULONG) fs_inode->ctime, (ULONG) dtime,
				       (ULONG) fs_inode->mode, (int) fs_inode->nlink, link,
				       (ULONG) fs_data->size);
				
	 	           	fs->file_walk(fs, fs_inode, fs_data->type, fs_data->id,
				     FS_FLAG_CONT | FS_FLAG_ALLOC | FS_FLAG_UNALLOC | FS_FLAG_NOABORT,
	 	              		print_addr, (char *)&ptr);
	
				print_blocks(inum, fs_data->type, fs_data->id, &run);
	 	       	}
			fs_data = fs_data->next;
		}
	}
	else {

	  //printf("INSERT INTO inode_%s VALUES('%lu','0','0','%c','%d','%d','%lu',"
	  printf("INSERT INTO inode_%s VALUES('%lu','%c','%d','%d','%lu'," \
		 "'%lu','%lu',%lu,'%lo','%d','%s','%lu');\n",
		 tbl_name,
		 (ULONG) inum, (flags & FS_FLAG_ALLOC) ? 'a' : 'f',
		 (int) fs_inode->uid, (int) fs_inode->gid,
		 (ULONG) fs_inode->mtime, (ULONG) fs_inode->atime,
		 (ULONG) fs_inode->ctime, (ULONG) dtime,
		 (ULONG) fs_inode->mode, (int) fs_inode->nlink, link,
		 (ULONG) fs_inode->size);
	  
	  if(fs_inode->size > 0) {
	    RUN run;
	    run.addr = -1;
	    run.next = NULL;
	    RUN *ptr = &run;   
	    
	    fs->file_walk(fs, fs_inode, 0, 0, 
			  (FS_FLAG_AONLY | FS_FLAG_CONT | FS_FLAG_META |
			   FS_FLAG_ALLOC | FS_FLAG_UNALLOC| FS_FLAG_NOABORT ),
			  print_addr, (char *)&ptr);
	    
	    print_blocks(inum, 0, 0, &run);
	  }
	}

	// adjust times back again
	if (sec_skew != 0) {
	  fs_inode->mtime += sec_skew;
	  fs_inode->atime += sec_skew;
	  fs_inode->ctime += sec_skew;
	}
	free(link);
	return WALK_CONT;
}

/* 
 * call back action function for file_walk
 * simply populate a linked list
 */
static u_int8_t
print_addr (FS_INFO *fs, DADDR_T addr, char *buf,
	    int size, int flags, char *ptr)
{
  // skip if no data
  if(size > 0) {
    /* indirect block pointer, dont add to list */
    if (flags & FS_FLAG_META) {
      return WALK_CONT;
    }
    
    RUN **run = (RUN **)ptr;
    
    //	if( ((fs->ftype & FSMASK) == NTFS_TYPE) && ((flags & FS_FLAG_CONT) == 0)) {
    if(((fs->ftype & FSMASK) == NTFS_TYPE) && !(flags & FS_FLAG_ALLOC) && !(flags & FS_FLAG_UNALLOC)) {
      // we have resident ntfs data
      //printf("resident found, size: %d\n", size);
      //printf("INSERT INTO resident_%s values('%lu','%d','%d','", 
      printf("INSERT INTO resident_%s values('%lu-%d-%d','", 
	     tbl_name, (ULONG)(*run)->inum, (*run)->type, (*run)->id);
      print_sql_string(stdout, buf, size);
      printf("');\n");
    }
    else {
      (*run)->addr = (unsigned long)addr;
      (*run)->next = (RUN *)malloc(sizeof(RUN));
      if(!(*run)->next) {
	printf("unable to allocate memory\n");
	exit(0);
      }
      (*run)->next->addr = -1;
      (*run)->next->next = NULL;
      *run = (*run)->next;
    }
  }
  return WALK_CONT;
}

//Function prints the string given in ptr as an SQL escaped string sequence. 
void print_sql_string(FILE *fh, const unsigned char *ptr, int length) {
  int i;
  for(i=0;i<length;i++) {
    switch(*(ptr+i)) {
    case 0:
      fprintf(fh,"\\0");
      break;
    case '\'':
      fprintf(fh,"\\'");
      break;
    case '\"':
      fprintf(fh,"\\\"");
      break;
    case '\n':
      fprintf(fh,"\\n");
      break;
    case '\\':
      fprintf(fh,"\\\\");
      break;
    default:
      fprintf(fh,"%c",*(ptr+i));
    };
  };
}

/* 
 * print file runs
 */
void print_blocks (INUM_T inum, int type, int id, RUN *run) {
	unsigned long index = 0, count = 1;
	unsigned long start_block = 0;
	RUN *ptr = run;
	RUN *ptr2 = run->next;

//	printf("INSERT INTO block_%s VALUES\n", tbl_name);

	while(ptr->addr != -1) {
	  if((ptr->next != NULL) && (ptr->next->addr == ptr->addr + 1)) {
	    if(start_block == 0) {
	      start_block = ptr->addr;
	    }
	    count++;
	  }
	  else {
	    if(start_block != 0) {
	      //printf("('%lu','%i','%i','%lu','%lu','%lu'),\n", (ULONG)inum, type, id, index, start_block, count);
	      //printf("INSERT INTO block_%s VALUES('%lu','%i','%i','%lu','%lu','%lu');\n", tbl_name, (ULONG)inum, type, id, index, start_block, count);
	      printf("INSERT INTO block_%s VALUES('%lu-%i-%i','%lu','%lu','%lu');\n", tbl_name, (ULONG)inum, type, id, index, start_block, count);
	    }
	    else {
	      //printf("('%lu','%i','%i','%lu','%lu','%lu'),\n", (ULONG)inum, type, id, index, ptr->addr, count);
	      //printf("INSERT INTO block_%s VALUES('%lu','%i','%i','%lu','%lu','%lu');\n", tbl_name, (ULONG)inum, type, id, index, ptr->addr, count);
	      printf("INSERT INTO block_%s VALUES('%lu-%i-%i','%lu','%lu','%lu');\n", tbl_name, (ULONG)inum, type, id, index, ptr->addr, count);
	    }
	    start_block = 0;
	    index += count;
	    count = 1;
	  }
	  ptr = ptr->next;
	}
	
	//printf(";\n");
	
	// free memory
	RUN *old;
	while(ptr2 != NULL) {
	  old = ptr2;
	  ptr2 = ptr2->next;
	  free(old);
	}
	
}

void print_fsinfo(FS_INFO *fs) {
  
  // fill in the metadata about the filesystem
  printf("INSERT INTO meta_%s VALUES\n" \
	 "('fstype', '%s'),\n('first_inode','%lu'),\n" \
	 "('last_inode','%lu'),\n('root_inode','%lu'),\n" \
	 "('first_block','%lu'),\n('last_block','%lu'),\n" \
	 "('block_size','%lu');\n\n",
	 tbl_name, fstype, 
	 (ULONG)fs->first_inum, (ULONG)fs->last_inum, 
	 (ULONG)fs->root_inum, (ULONG)fs->start_block, 
	 (ULONG)fs->last_block, (ULONG)fs->block_size);
}

/* 
 * main
 */
int 
main(int argc, char **argv) 
{
	int f_flags = FS_FLAG_ALLOC | FS_FLAG_UNALLOC | FS_FLAG_RECURSE;
	int i_flags = 0;
	char ch;
	FS_INFO 	*fs;
	char *dbaction = NULL;
	extern int optind;
	char *io_subsys=NULL;
	IO_INFO *io=NULL;

	progname = argv[0];
	fstype = DEF_FSTYPE;

	i_flags |= ~0;

	while ((ch = getopt(argc, argv, "i:t:f:d:s:vVz:")) > 0) {
	  switch (ch) {
	  case '?':
	  default: 
	    usage(argv[0]);
	  case 'f':
	    fstype = optarg;
	    break;
	  case 'i':
	    io_subsys=optarg;
	    break;
	  case 't':
	    tbl_name = optarg;
	    break;
	  case 'd':  
	    dbaction=optarg;
	    break;
	  case 's':
	    sec_skew = atoi(optarg);
	    break;
	  case 'v':
	    verbose++;
	    logfp = stderr;
	    break;
	  case 'V':
	    print_version();
	    
	    exit(0);
	    break;
	  case 'z':
	    {
	      char envstr[32];
	      snprintf(envstr, 32, "TZ=%s", optarg);
	      if (0 != putenv(envstr)) {
		RAISE (E_GENERIC,NULL,"error setting environment");
	      }
	      
	      /* we should be checking this somehow */
	      tzset();
	    }
	    break;
	  }
	}

	if(!tbl_name) RAISE(E_GENERIC,NULL,"Table name not specified!!");

	/* create/drop tables */
	if(dbaction) {
	  if(!strcmp(dbaction, "create")) {
	    create_tables(tbl_name);
	    exit(0);
	  } else if(dbaction && !strcmp(dbaction, "drop")) {
	    drop_tables(tbl_name);
	    exit(0);
	  } else {
	    RAISE(E_GENERIC,NULL,"Only valid parameters to -d are create or drop");
	  };
	};
	
	if(optind == argc)
	  usage(argv[0]);

	/* User chose to set the io_subsystem */
	if(io_subsys) {
	  io=io_open(io_subsys);
	} else {
	  /* If the user did not specify a subsystem, we choose the standard one */
	  io=io_open("standard");
	};

	if(!io) {
	  RAISE(E_IOERROR,NULL,"Could not set io subsystem %s",io_subsys);
	};

	//Parse the rest of the args as options to the io filesystem:
	while(optind<argc) {
	  io_parse_options(io,argv[optind++]);
	};

	/* open image */
	fs = fs_open(io, fstype);
	if(!fs) RAISE(E_GENERIC,NULL,"Unable to open file system as %s",fstype);

	/* print filesystem info 
	 * This fills in the meta table
	 */
	print_fsinfo(fs);

	/* directory walk, fills in the file table
	 */
	//printf("INSERT INTO file_%s VALUES\n", tbl_name);
	fprintf(stderr, "Loading Directory Entries\n");
	fs->dent_walk(fs, fs->root_inum, f_flags, print_dent, (char *)0); 
	//printf(";\n");

	/* inode walk
	 * This fills in the inode table, the callback also calls file_walk
	 * which fills in the blocks table.
	 */
	inode_count = fs->first_inum;
	inode_total = fs->last_inum;
	fprintf(stderr, "Loading Inode Entries\n");
	fs->inode_walk(fs, fs->first_inum, fs->last_inum, i_flags, 
		       print_inode, (char *)0);
	/* close file */
	fs->close(fs);

	exit (0);
}

void create_tables(char *name) {
	/* create tables */
	printf("CREATE TABLE IF NOT EXISTS meta_%s (`name` VARCHAR(25) NOT NULL, `value` VARCHAR(25) NOT NULL);\n\n", name);

	// these fields are all joined in 'inode' now...
	//"	`inode` INT NOT NULL,\n"
	//"	`type` INT NOT NULL,\n"
	//"	`id` INT NOT NULL,\n"

	printf("CREATE TABLE IF NOT EXISTS inode_%s (\n" \
	"	`inode` VARCHAR(20) NOT NULL,\n" \
	"	`status` INT,\n" \
	"	`uid` INT,\n" \
	"	`gid` INT,\n" \
	"	`mtime` INT NOT NULL,\n" \
	"	`atime` INT NOT NULL,\n" \
	"	`ctime` INT NOT NULL,\n" \
	"	`dtime` INT,\n" \
	"	`mode` INT,\n" \
	"	`links` INT,\n" \
	"	`link` TEXT,\n" \
	"	`size` BIGINT NOT NULL);\n\n", name);;

	printf("CREATE TABLE IF NOT EXISTS file_%s (\n" \
	"	`inode` VARCHAR(20) NOT NULL,\n" \
	"	`mode` VARCHAR(3) NOT NULL,\n" \
	"	`status` VARCHAR(8) NOT NULL,\n" \
	"	`path` TEXT,\n" \
	"	`name` TEXT);\n\n", name);

	printf("CREATE TABLE IF NOT EXISTS block_%s (\n" \
	"	`inode` VARCHAR(20) NOT NULL,\n" \
	"	`index` INT NOT NULL,\n" \
	"	`block` BIGINT NOT NULL,\n" \
	"	`count` INT NOT NULL);\n\n", name);

	printf("CREATE TABLE IF NOT EXISTS resident_%s (\n" \
	"	`inode` VARCHAR(20) NOT NULL,\n" \
	"	`data` TEXT);\n\n", name);
}

void drop_tables(char *name) {
	/* drop tables */
	printf("DROP TABLE IF EXISTS meta_%s, inode_%s, file_%s, block_%s, resident_%s;\n", name, name, name, name, name);
}
