/*++
* NAME
*	fs_io 3
* SUMMARY
*	file system I/O routines
* SYNOPSIS
*	#include "fstools.h"
*
*	void	fs_read_block(fs, buf, len, addr, comment)
*	FS_INFO *fs;
*	FS_BUF	*buf;
*	int	len;
*	DADDR_T	addr;
*	const char *comment;
*
*	void	fs_read_random(fs, buf, len, offs, comment)
*	FS_INFO *fs;
*	char	*buf;
*	int	len;
*	OFF_T	offs;
*	const char *comment;
* DESCRIPTION
*	fs_read_block() reads a block of data from the named file system.
*
*	fs_read_random() reads from an arbitrary position.
*
*	Arguments:
* .IP fs
*	File system handle.
* .IP buf
*	Result buffer pointer.
* .IP len
*	Amount of bytes to be read.
* .IP addr
*	Block number.
* .IP fd
*	Open device file.
* .IP offs
*	Byte offset.
* .IP comment
*	Text logged with verbose logging.
* LICENSE
*	This software is distributed under the IBM Public License.
* AUTHOR(S)
*	Wietse Venema
*	IBM T.J. Watson Research
*	P.O. Box 704
*	Yorktown Heights, NY 10598, USA
*                Michael Cohen (scudette@users.sourceforge.net) 
*                        - Implemented pluggable IO subsystem
*/

#include "fs_tools.h"
#include "error.h"
#include "fs_types.h"
#include "fs_io.h"
#include <assert.h>
#include "sgzlib.h"
#include "libevf.h"

/*******************************************
 *          Standard IO Subsystem.
 *          Sleuthkit default.
 *******************************************/
struct IO_INFO_STD {
  IO_INFO io;
  char *name;
  int fd;
};

typedef struct IO_INFO_STD IO_INFO_STD;

/* This is a generic constructor that can be called from all other
   objects */
IO_INFO *io_constructor(IO_INFO *class) {
  IO_INFO *io=(IO_INFO *)calloc(1,class->size);
  if(!io) error("Could not malloc\n");
  
  //Instantiate the class in the new object:
  memcpy(io,class,sizeof(*class));
  return(io);
};

/* fs_read_random - random-access read */
int    std_read_random(IO_INFO *self, char *buf, int len, OFF_T offs,
		               const char *comment)
{
    char   *myname = "fs_read_random";
    int     count;
    IO_INFO_STD *io=(IO_INFO_STD *)self;

    if(!self->ready && self->open(self)<0) return(-1);

    if (verbose)
	fprintf(logfp, "%s: read offs %.0f len %d (%s)\n",
		myname, (double) offs, len, comment);

    if (LSEEK(io->fd, offs, SEEK_SET) != offs)
      RAISE(E_IOERROR,NULL,"read random seek error: offset %llu:", 
	    (ULLONG) offs);

    if ((count = read(io->fd, buf, len)) != len)
      RAISE(E_IOERROR,NULL,"read random read error (%d@%llu):", 
	    len, (ULLONG) offs);

    //    if(fs) fs->seek_pos = offs + len;
    return(len);
}

void std_help(void) {
  printf("No parameters required");
};

int std_initialiser(IO_INFO *io,IO_OPT *args) {
  IO_INFO_STD *self=(IO_INFO_STD *)io;

  // If we only get an option with no value, then we assume its the
  // filename we need to open:
  for(;args;args=args->next) { 
    if(!args->value || strlen(args->value)==0) {
      self->name=strdup(args->option);
    } else if(!strcmp(args->option,"filename")) {
      self->name=strdup(args->value);
    };
  };
  return(0);
};

int std_open(IO_INFO *self) {
  IO_INFO_STD *io=(IO_INFO_STD *)self;
  if(!io->name) {
    RAISE(E_IOERROR,NULL,"No filename set!!!");
  };
  io->fd=open(io->name,O_RDONLY);
  if(io->fd<0) {
    RAISE(E_IOERROR,NULL,"Unable to open file %s",io->name);
  };
  self->ready=1;
  return(io->fd);
};

int std_close(IO_INFO *io) {
  IO_INFO_STD *self=(IO_INFO_STD *)io;
  return (close(self->fd));
};

/***************************************
 *     Advanced IO Subsystem 
 *
 *   This subsystem can handle the dd image being split across
 *   multiple files, as well as having the partition image located
 *   within a larger file (for example the disk dd image)
 ***************************************/
struct adv_split_file {
  const char *name;
  int fd;
  // These correspond to the offsets for this file
  long long int start_offset;
  long long int end_offset;
  struct adv_split_file *next;
};

struct IO_INFO_ADV {
  IO_INFO io;
  //Stores the files belonging to this dd image set
  struct adv_split_file *file_list;
  unsigned long long offset;
};

typedef struct IO_INFO_ADV IO_INFO_ADV;

/* fs_read_random - random-access read */
int    adv_read_random(IO_INFO *self, char *buf, int len, OFF_T offs,
		               const char *comment)
{
    IO_INFO_ADV *io=(IO_INFO_ADV *)self;
    unsigned long long int     count,available,buflen=len;
    struct adv_split_file *file;

    if(!self->ready)
      self->open(self);

    // Correct the offset as per the user preference:
    offs+=io->offset;

    //Find the file in the split list where the seek will end up in.
    for(file=io->file_list; file; file=file->next) {
      if((file->start_offset<=offs) && (offs<file->end_offset))
	break;
    };

    while(len>0) {
      if(!file)
	RAISE(E_IOERROR,NULL,"read random seek error: offset %llu:", 
	      (ULLONG) offs);
      
      // Now find out how much we can read from this file:
      available= file->end_offset - offs;
      
      //Seek to the right spot:
      if (LSEEK(file->fd, offs - file->start_offset, SEEK_SET) != offs - file->start_offset)
	RAISE(E_IOERROR,NULL,"read block lseek error (%llu):", 
	      (ULLONG) offs);
      
      // If we only want to read a smaller number than whatever is
      // available:
      if(len<available) available=len;
      
      // Read that much into buf:
      if ((count = read(file->fd, buf, available)) != available)
	RAISE(E_IOERROR,NULL,"read random read error (%d@%llu): ", 
	      len, (ULLONG) offs);
      
      //Advance to the next file in the sequence:
      len-=available;
      offs+=available;
      buf+=available;
      file=file->next;
    }
    //Return the total count of bytes read
    return(buflen);
}

void adv_help(void) {
  printf("Advanced io subsystem options\n\n");
  printf("\toffset=bytes\t\tNumber of bytes to seek to in the image file. Useful if there is some extra data at the start of the dd image (e.g. partition table/other partitions)\n");
  printf("\tfile=filename\t\tFilename to use for split files. If your dd image is split across many files, specify this parameter in the order required as many times as needed for seamless integration\n");
  printf("\tA single word without an = sign represents a filename to use\n");
};

int adv_add_file_to_split_list(IO_INFO_ADV *io,const char *name) {
  /* Add file to our split files list: */
  struct adv_split_file *file,*tmp;
  long long int max_offset=0;
  long long int file_size=0;
  
  file=(struct adv_split_file *)malloc(sizeof(struct adv_split_file));
  if(!file) RAISE(E_NOMEMORY,NULL,"Cant Malloc");

  file->next=NULL;  
  file->name=name;
  file->fd = open(name,O_RDONLY);
  if(file->fd<0) {
    free(file);

    RAISE(E_IOERROR,NULL,"Could not open file %s",name);
  };
  
  /* Find out the maximum offset already stored in the file list, as
     well as the end of the list */
  if(io->file_list) {
    for(tmp=io->file_list;tmp->next;tmp=tmp->next);
    tmp->next=file;
    max_offset=tmp->end_offset;
  } else {
    io->file_list=file;
  };
  
  /* Find out the size of the file: */
  file_size=LSEEK(file->fd,0,SEEK_END);
  if(file_size<0){
    free(file);

    RAISE(E_IOERROR,NULL,"Unable to seek on file %s",file->name);
  };
  
      /* Set the bounds of this file */
  file->start_offset=max_offset;
  file->end_offset=file->start_offset+file_size;
  
  return(file->fd);
};

int adv_initialiser(IO_INFO *self,IO_OPT *args) {
  IO_OPT *i;
  IO_INFO_ADV *io=(IO_INFO_ADV *)self;

  for(i=args;i;i=i->next) {
    //If we are not given an option=value pair, we interpret the whole
    //thing as a filename:
    if(args->option && (!args->value || strlen(args->value)==0)) {
      adv_add_file_to_split_list(io,i->option);
      continue;
    }else if(!strncasecmp(i->option,"file",strlen("file"))) {
      adv_add_file_to_split_list(io,i->value);
      fprintf(stderr, "Set file to read from as %s\n",i->value);
      continue;
    } else if(!strncasecmp(i->option,"offset",strlen("offset"))) {
      io->offset=atoll(i->value);
      fprintf(stderr, "setting offset to %llu\n",io->offset);
      continue;
    };

    // If we get here we did not recognise this option, lets give the
    // user a helpful message:
    RAISE(E_GENERIC,NULL,"option %s not recognised",i->option);
  };
  return(0);
};

int adv_open(IO_INFO *self) {
  //We dont do any openning here, since we open the files when we add
  //the options in...
  self->ready=1;
  return(1);
};

int adv_close(IO_INFO *self) {
  IO_INFO_ADV *io=(IO_INFO_ADV *)self;
  struct adv_split_file *file;
  
  for(file=io->file_list;file;file=file->next) 
    close(file->fd);
  return (1);
};

/************************************************
 *                       sgzip subsystem
 *
 *   This subsystem allows sleuthkit to read compressed sgzip image
 *   files. sgzip is a seekable compressed file format that can be
 *   generated by the use of the sgzip utility. See sgzlib.h for a
 *   full explanatation of the format.
 ************************************************/
struct IO_INFO_SGZ {
  IO_INFO io;
  struct sgzip_header *header;
  int fd;
  char *name;
  unsigned long long int *index;
  unsigned long long int offset;
};

typedef struct IO_INFO_SGZ IO_INFO_SGZ;

int sgz_read_random(IO_INFO *self, char *buf, int len, OFF_T offs,
		               const char *comment)
{
  IO_INFO_SGZ *io=(IO_INFO_SGZ *)self;

  if(!self->ready && self->open(self)<0) return(-1);

  sgzip_read_random(buf,len,offs+io->offset,io->fd,io->index,io->header);
  return(len);
}

void sgz_help(void) {
  printf("sgzip subsystem options\n\n");
  printf("\toffset=bytes\t\tNumber of bytes to seek to in the (uncompressed) image file. Useful if there is some extra data at the start of the dd image (e.g. partition table/other partitions\n");
};

int sgz_open(IO_INFO *self) {
  IO_INFO_SGZ *io=(IO_INFO_SGZ *)self;
  
  if(!io->name) {
    RAISE(E_IOERROR,NULL,"No filename set!");
  };

  io->fd=open(io->name,O_RDONLY);
  if(io->fd<0) {
    RAISE(E_IOERROR,NULL,"Could not open file %s",io->name);
  };

  io->header=sgzip_read_header(io->fd);
  if(!io->header) {
    RAISE(E_IOERROR,NULL,"%s is not a sgz file",io->name);
  };

  io->index=sgzip_read_index(io->fd,io->header);
  if(!io->index) {
    io->header=sgzip_read_header(io->fd);
    fprintf(stderr, "You may consider rebuilding the index on this file to speed things up, falling back to non-indexed method\n");
    io->index=sgzip_calculate_index_from_stream(io->fd,io->header);
  };
  self->ready=1;
  return(io->fd);
};

int sgz_initialiser(IO_INFO *self,IO_OPT *args) {
  IO_OPT *i;
  IO_INFO_SGZ *io=(IO_INFO_SGZ *)self;

  for(i=args;i;i=i->next) {
    //If we are not given an option=value pair, we interpret the whole
    //thing as a filename:
    if(args->option && (!args->value || strlen(args->value)==0)) {
      io->name=strdup(args->option);
      continue;
    }else if(!strncasecmp(i->option,"file",strlen("file"))) {
      io->name=strdup(args->value);
      fprintf(stderr, "Set file to read from as %s\n",i->value);
      continue;
    } else if(!strncasecmp(i->option,"offset",strlen("offset"))) {
      io->offset=atoll(i->value);
      fprintf(stderr, "setting offset to %llu\n",io->offset);
      continue;
    };

    // If we get here we did not recognise this option, raise an
    // error:
    RAISE(E_GENERIC,NULL,"option %s not recognised",i->option);
  };
  return(0);
};

int sgz_close(IO_INFO *self) {
  IO_INFO_SGZ *io=(IO_INFO_SGZ *)self;
  free(io->index);
  free(io->header);
  return (close(io->fd));
};

/************************************************
 *  Expert Witness Compression Format (EWF) subsystem
 *
 *   This subsystem allows processing of image files taken with the
 *   EWF format. EWF is a forensic compression format which is a
 *   "quasi-proprietary format", primarily used by Encase (tm)
 *   (http://www.guidancesoftware.com/) but also used by other leading
 *   commercial products.
 *
 *   This implementation is a clean room implementation based on the
 *   specifications on http://www.asrdata.com/SMART/whitepaper.html.
 *   This implementation is available under the GPL.
 *
 *   This subsystem allows interoperability with image files acquired
 *   using Encase, FTK and SMART.
 *
 *************************************************/
struct IO_INFO_EWF {
  IO_INFO io;
  struct evf_file_header *file_header;
  struct offset_table offsets;
  //This is the offset into the image. Its usually zero unless the
  //entire HDD was acquired.
  unsigned long long int offs;
  //Should we ignore errors?
  int force;
};

typedef struct IO_INFO_EWF IO_INFO_EWF;

void ewf_help(void) {
  printf("An Expert Witness IO subsystem\n\n"); 
  printf("\toffset=bytes\t\tNumber of bytes to seek to in the (uncompressed) image file. Useful if there is some extra data at the start of the dd image (e.g. partition table/other partitions\n");
  printf("\tfilename=file.e0?\t\tFilename to use for split files. If your dd image is split across many files, specify this parameter in the order required as many times as needed for seamless integration\n");
  printf("\tA single word without an = sign represents a filename to use\n");
};

int ewf_read_random(IO_INFO *self, char *buf,int len, OFF_T offs,
		    const char *comment) {
  IO_INFO_EWF *io=(IO_INFO_EWF *)self;
  
  //If we are not ready, we open the files
  if(!self->ready) self->open(self);
  
  //Read the data - Note we dont need to handle exceptions here, just
  //let them bubble upwards...
  evf_read_random(buf,len,offs+io->offs,&(io->offsets));
  
  return(len);
};

/* Opens and Adds additional files */
int ewf_add_file(IO_INFO_EWF *self,char *filename) {
  int fd;
  struct evf_file_header *file_header;

  fd=open(filename,O_RDONLY);
  if(fd<0) {
    RAISE(E_IOERROR,NULL,"Could not open %s",filename);
  };

  file_header=evf_read_header(fd);
  
  //Grow the files array so we can fit the segment in it:
  if(self->offsets.max_segment < file_header->segment) {
    //Amount of additional memory we will need
    int additional_memory=file_header->segment +1;
    
    self->offsets.files=realloc(self->offsets.files,additional_memory*sizeof(*self->offsets.files));
    if(!self->offsets.files) RAISE(E_NOMEMORY,NULL,"Malloc");
    
    //Clear off the newly initialised memory
    memset(self->offsets.files + self->offsets.max_segment + 1, -1 , (additional_memory - self->offsets.max_segment - 1)*sizeof(*self->offsets.files));
    
    //Adjust the maximum size of the array
    self->offsets.max_segment = file_header->segment;
  };
  
  //Store the fd in the correct spot within the files array. So we end
  //up with all the segments ordered regardless of which order we
  //opened them in.
  if(self->offsets.files[file_header->segment]>0) RAISE(E_GENERIC,NULL,"A segment (%s) is specified more than once",filename);
  
  self->offsets.files[file_header->segment]=fd;
  return(0);
};

int ewf_initialiser(IO_INFO *self,IO_OPT *args) {
  IO_OPT *i;
  IO_INFO_EWF *io=(IO_INFO_EWF *)self;

  //Initialise the offsets table the first time we use it:
  if(!io->offsets.files) {
    io->offsets.max_chunk=0;
    io->offsets.max_segment=0;
    //Allocate one array entry and realloc the rest when we need it
    io->offsets.files=(unsigned int *)malloc(sizeof(io->offsets.files)*2);
    io->offsets.fd=NULL;
    io->offsets.offset=NULL;
    io->offsets.section_list=NULL;
    *(io->offsets.files)=-1;
  };
  
  for(i=args;i;i=i->next) {
    if(!strncasecmp(i->option,"help",strlen("help"))) {
      self->help();
      continue;

    //If we are not given an option=value pair, we interpret the whole
    //thing as a filename:
    } else if(args->option && (!args->value || strlen(args->value)==0)) {
      return(ewf_add_file(io,strdup(args->option)));

    }else if(!strncasecmp(i->option,"file",strlen("file"))) {
      return(ewf_add_file(io,strdup(args->value)));

    } else if(!strncasecmp(i->option,"offset",strlen("offset"))) {
      io->offs=atoll(i->value);
      continue;
    };

    // If we get here we did not recognise this option, raise an error:
    RAISE(E_GENERIC,NULL,"option %s not recognised",i->option);
  };
  return(0);
};

int ewf_open(IO_INFO *self) {
  struct evf_section_header *section=NULL;
  IO_INFO_EWF *io=(IO_INFO_EWF *)self;
  int i;
  char tmp[255];

  //When we get here we should have all the files opened, and ready to
  //go. So we just check for consistency that we do not have any files
  //that were not specified:
  
  if(!io->offsets.files || io->offsets.max_segment<1) {
    RAISE(E_IOERROR,NULL,"No Files given");
  };

  for(i=1;i<=io->offsets.max_segment;i++) {
    if(io->offsets.files[i]<0) {
      RAISE(E_IOERROR,NULL,"Missing a segment file for segment %u",i);
    };
    
    //Now process each file in order until we build the whole index
    for(section=evf_read_section(io->offsets.files[i]);
	section->size>0;
	section=evf_read_section(io->offsets.files[i])) {
      
      //This will update offsets.fds and offsets.offset
      process_section(section,i,&(io->offsets));
      
      //Go to the next section
      if(lseek(io->offsets.files[i],section->next,SEEK_SET)<0) {
	RAISE(E_IOERROR,NULL,"Could not seek\n");
      };

      strncpy(tmp,section->type,250);
      free(section);
    };
  };

  //Check to see if we are done?
  if(strcasecmp(tmp,"done") && io->force) 
    RAISE(E_IOERROR,NULL,"No ending section, Cant find the last segment file");

  self->ready=1;  
  return(1);
}

int ewf_close(IO_INFO *self) {
  IO_INFO_EWF *io=(IO_INFO_EWF *)self;
  
  free(io->offsets.files);
  free(io->offsets.fd);
  free(io->offsets.offset);
  free(io->offsets.size);
  free(io);
  return(1);
};

/************************************************
 *                     RAID subsystem
 *
 *    This subsystem allows the sleuthkit to read files off a raid set
 *    given a raid reconstruction map.
 *
 * Not implemented yet
 ************************************************/

/* These serve as classes (i.e. templates which each object
   instantiates). */
static IO_INFO subsystems[] ={
  { "standard","Standard Sleuthkit IO Subsystem",sizeof(IO_INFO_STD),&io_constructor,&free, 
    &std_help, &std_initialiser,  &std_read_random,  &std_open,  &std_close,0},

  { "advanced","Advanced Sleuthkit IO Subsystem",sizeof(IO_INFO_ADV),&io_constructor,&free,
    &adv_help, &adv_initialiser, &adv_read_random, &adv_open, &adv_close,0},

  { "sgzip","Seekable Gzip format",sizeof(IO_INFO_SGZ),&io_constructor,&free,  
    &sgz_help, &sgz_initialiser, &sgz_read_random, &sgz_open, &sgz_close,0},
  
  { "ewf","Expert Witness Compression format",sizeof(IO_INFO_EWF),&io_constructor,&free,  
    &ewf_help, &ewf_initialiser, &ewf_read_random, &ewf_open, &ewf_close,0},

  //Sentinel
  { NULL, NULL, 0, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL }
  };

// Constructor for io_opts
IO_OPT *new_io_opt() {
  IO_OPT *result;

  result=(IO_OPT *)malloc(sizeof(IO_OPT));
  assert(result);
  result->option=NULL;
  result->value=NULL;
  result->next=NULL;
  return(result);
};

void io_parse_options(IO_INFO *io_obj,char *opts) {
  char *x,*y,*z;
  char *temp=strdup(opts);
  IO_OPT *io=NULL;
  IO_OPT *options=NULL;

  z=temp;
  while(1) {
    // Find the next comma:
    y=index(z,',');
    if (y) *y='\0';
    
    //Now find the = sign
    x=index(z,'=');
    
    if(x) {
      *x='\0';
      x++;
    };

    //Go to the end of the opts list
    if(options) {
      for(io=options;io->next;io=io->next);
      io->next=new_io_opt();
      io=io->next;
    } else {
      options=new_io_opt();
      io=options;
    };
    io->option=z;
    io->value=x;
    if(!y) break;
    z=y+1;
  };
  //Call the initialiser with those args:
  io_obj->initialise(io_obj,options);
};

/* Selects the requested subsystem.

Returns 0 for success, -1 for failure
*/
IO_INFO *io_open(char *name) {
  IO_INFO *io=NULL;
  int i=0;
  for(i=0;subsystems[i].name;i++) {
    if(!strncasecmp(subsystems[i].name,name,strlen(subsystems[i].name))) {
      io=&(subsystems[i]);
    };
  };

  if(io) {
    //Create a new instance
    io=io->constructor(io);
    return(io);
  };

  RAISE(E_IOERROR,NULL,"No such IO subsystem: %s",name);
  return(NULL);
};

int io_close(IO_INFO *self) {
  return(self->close(self));
};

void io_help(char *name) {
  int i;

  for(i=0;subsystems[i].name;i++) {
    if(!strncasecmp(subsystems[i].name,name,strlen(subsystems[i].name))) {
      subsystems[i].help();
      return;
    };
  };

  printf("Available Subsystems:\n\n");
  /* Print a list of subsystems to let the user know which ones are
     available: */
  for(i=0;subsystems[i].name;i++) {
    printf("\t%s - %s\n",subsystems[i].name,subsystems[i].description);
  };
};

/* This is the generic read block function which all the fs call */
void  fs_read_block(FS_INFO *fs, FS_BUF *buf, int len, DADDR_T addr,
		    const char *comment)
{
  char   *myname = "fs_read_block";
  OFF_T   offs;
  
  /*
   * Sanity checks.
   */
  
  if (len % fs->dev_bsize)
    RAISE(E_IOERROR,NULL,"%s: Block Read Request with length (%d) not a multiple of %d (%s)", 
	  myname, len, fs->dev_bsize, comment);
  
  if (len > buf->size)
    RAISE(E_IOERROR,NULL,"%s: Buffer length is too short for read (%d > %d) (%s)", 
	  myname, len, buf->size, comment);
  
  if (addr > fs->last_block)
    RAISE(E_IOERROR,NULL,"%s: Read Error (address is too large for image %lu) (%s)", 
	  myname, (ULONG)addr, comment);
  
  buf->addr = addr;
  offs = (OFF_T) addr *fs->block_size;
  
  if (verbose)
    fprintf(logfp, "%s: read block %lu offs %llu len %d (%s)\n",
	    myname, (ULONG) addr, (ULLONG) offs, len, comment);
  
  // Call the random read method to do the actual reading...
  fs->io->read_random(fs->io,buf->data,len,offs,comment);
  
  fs->seek_pos = offs + len;
  buf->used = len;
  return;
}
