# ******************************************************
# Copyright 2004: Commonwealth of Australia.
#
# Developed by the Computer Network Vulnerability Team,
# Information Security Group.
# Department of Defence.
#
# Michael Cohen <scudette@users.sourceforge.net>
# David Collett <daveco@users.sourceforge.net>
#
# ******************************************************
#  Version: FLAG 0.4 (12-02-2004)
# ******************************************************
#
# * This program is free software; you can redistribute it and/or
# * modify it under the terms of the GNU General Public License
# * as published by the Free Software Foundation; either version 2
# * of the License, or (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ******************************************************

""" Flag module for performing structured disk forensics """
import pyflag.Reports as Reports
import pyflag.FlagFramework as FlagFramework
import pyflag.conf
config=pyflag.conf.ConfObject()
import os,os.path,time,re
import pyflag.Sleuthkit as Sleuthkit
import pyflag.Graph as Graph

description = "Disk Forensics"
order=30

import pyflag.IO as IO

class BrowseFS(Reports.report):
    """ Report to browse the filesystem"""
    parameters = {'fsimage':'fsimage'}
    hidden = False
    name = "Browse Filesystem"
    description = "Display filesystem in a browsable format"
    
    def display(self,query,result):
        result.heading("Browsing Filsystem in image %s" % query['fsimage'])

        # lookup the iosource for this fsimage
        iofd = IO.open(query['case'], query['fsimage'])
        fsfd = Sleuthkit.DDFS( query["case"], query["fsimage"], iofd)
        
        branch = ['']
        new_query = result.make_link(query, '')
        if (query.has_key("open_tree") and query['open_tree'] != '/'):
            br = query['open_tree']
        else:
            br = ''

        if not fsfd.isdir(br):
            br = br[:br.rindex('/')]

        result.para("Inspecting branch %s" % br)
            
        def tree_cb(branch):
            path ='/'.join(branch)  + '/'
            fsfd = Sleuthkit.DDFS( query["case"], query["fsimage"], iofd)

            a = fsfd.dent_walk(path)
            for i in a:
                if i['mode']=="d/d":
                    link = self.ui(result)
                    link.link(i['name'],new_query,open_tree="%s%s" %(path,i['name']), mode='table', where_Filename="%s%s" %(path,i['name']), order='Filename')# ,__mark__="%s%s" %(path,i['name']))
                    yield(([i['name'],link,'branch']))

        def DeletedIcon(value):
            """ Callback for deleted items """
            tmp=self.ui()
            if value=='alloc':
                tmp.icon("yes.png")
            elif value=='deleted':
                tmp.icon("no.png")
            else:
                tmp.icon("question.png")

            return tmp

        try:
            # tabular view
            if query['mode'] == 'table':
                result.table(
                    columns=['f.inode','concat(path,name)','f.status','size','from_unixtime(mtime)','from_unixtime(atime)','from_unixtime(ctime)'],
                    names=('Inode','Filename','Del','File Size','Last Modified','Last Accessed','Created'),
                    callbacks={'Del':DeletedIcon},
                    table='file_%s as f, inode_%s as i' % (fsfd.table,fsfd.table),
                    where="f.inode=i.inode and f.mode like 'r/%%'",
                    case=query['case'],
                    links=[ FlagFramework.query_type((),case=query['case'],family=query['family'],report='ViewFile', fsimage=query['fsimage'],__target__='inode', inode="%s"), FlagFramework.query_type((),case=query['case'],family=query['family'],report='BrowseFS', fsimage=query['fsimage'],__target__='open_tree',open_tree="%s") ]
                    )
        except KeyError:
            # default to tree view
            query['order']='Filename'
            tmp = self.ui(result)            
            tmp.table(
                columns=['f.inode','name','f.status','size', 'from_unixtime(mtime)'],
                names=('Inode','Filename','Del','File Size','Last Modified'),
                callbacks={'Del':DeletedIcon},
                table='file_%s as f, inode_%s as i' % (fsfd.table,fsfd.table),
                where="f.inode=i.inode and path=%r and f.mode='r/r'" % (br + '/'),
                case=query['case'],
                links=[ FlagFramework.query_type((),case=query['case'],family=query['family'],report='ViewFile', fsimage=query['fsimage'],__target__='inode', inode="%s")]
                )
        
            tmp2 = self.ui(result)
            tmp2.tree(cb = tree_cb,branch = branch )
            result.start_table()
            result.row(tmp2, tmp, valign="top")
            result.end_table()
            result.next = tmp.next
            result.previous = tmp.previous
            result.pageno = tmp.pageno
            return result

    def form(self,query,result):
        try:
            result.case_selector()
            if query['case']!=config.FLAGDB:
               result.meta_selector(case=query['case'],property='fsimage')
        except KeyError:
            return result

class ViewFile(Reports.report):
    """ Report to browse the filesystem """
    parameters = {'fsimage':'fsimage','inode':'sqlsafe'}
    hidden = True
    name = "View File Contents"
    description = "Display the contents of a file"
    
    def display(self,query,result):
        new_q = result.make_link(query, '')
        if not query.has_key('limit'): query['limit']= 0

        # retrieve the iosource for this fsimage
        iofd = IO.open(query['case'],query['fsimage'])
        fsfd = Sleuthkit.DDFS( query["case"], query["fsimage"], iofd)
        fd = fsfd.open(inode=query['inode'])
        
        image = Graph.FileDump(fd)
        #How big is this file?
        i=fsfd.istat(inode=query['inode'])
        filesize=i['size']
        
        #Add the filename into the headers:
        path=fsfd.lookup(inode=query['inode'])
        if not path: raise IOError("No path for Inode %s" % query['inode'])
        path,name=os.path.split(path)
        image.headers=[("Content-Disposition","attachment; filename=%s" % name),
                       ("Content-Length",filesize)]
        
        result.heading("Viewing file in inode %s" % (query['inode']))
        try:
            result.text("Classified as %s by magic" % image.GetMagic())
        except IOError:
            result.text("Unable to classify file, no blocks")
            image = None

        def download(query):
            """ Used for dumping the entire file into the browser """
            if image:
                result.result=image.display()
                result.type=image.GetContentType()
                result.headers=image.headers
                result.binary=True
            return None

        def hexdump(query):
            """ Show the hexdump for the file """
            out=self.ui()
            if image:
                try:
                    max=config.MAX_DATA_DUMP_SIZE
                except AttributeError:
                    max=1024

                #Set limits for the dump
                try:
                    limit=int(query['hexlimit'])
                except KeyError:
                    limit=0
                dump = FlagFramework.HexDump(image.display(),out)
                dump.dump(offset=limit,limit=max)

                #Do the navbar
                result.next=limit+max
                if result.next>fd.size:
                    result.next=None
                result.previous=limit-max
                if result.previous<0:
                    if limit>0:
                        result.previous = 0
                    else:
                        result.previous=None
                result.pageno=limit/max
                result.nav_query=query.clone()
                result.nav_query['__target__']='hexlimit'
            else:
                out.text("No Data Available")

            return out

        def strings(query):
            """ Draw the strings in a file """
            str = pyflag.Strings.StringExtracter(fd)
            try:
                offset=query['stroffset']
                if offset.startswith("!"):
                    ## We search backwards for the correct offset
                    offset=str.find_offset_prior(int(offset[1:]),config.PAGESIZE-1)
            except KeyError:
                offset=0

            q=query.clone()
            del q['mode']
            del q['hexlimit']
            
            output=self.ui()
            output.start_table()
            row_number=0
            file_offset=offset
            try:
                for i in str.extract_from_offset(int(offset)):
                    row_number+=1
                    if row_number>config.PAGESIZE: break
                    file_offset=i[0]
                    tmp_link=self.ui()
                    tmp_link.link("0x%x (%s)" % (file_offset,file_offset),q,mode="HexDump",hexlimit=file_offset)
          
                    tmp_string=self.ui()
                    tmp_string.text(i[1],color="red",sanitise="full")
                    output.row(tmp_link,tmp_string,valign="top")

            except IOError:
                pass
            
            result.nav_query=query.clone()
            result.nav_query['__target__']='stroffset'
            result.next=file_offset
            if row_number<config.PAGESIZE: result.next=None
            result.previous="!%s" % offset
            result.pageno=offset

            return output

        def stats(query):
            """ Show statistics about the file """
            result=self.ui()
            istat = fsfd.istat(inode=query['inode'])
            left=self.ui()
            left.row("filename:",'',"%s/%s"%(path,name))
            for k,v in istat.iteritems():
                left.row('%s:' % k,'',v)
            left.end_table()

            if image:
                right=self.ui()
                right.image(image,width=200)
                result.start_table(width="100%")
                result.row(left,right,valign='top',align="left")
            else:
                result.join(left)
            return result

        result.notebook(
            names=["Statistics","HexDump","Download","Strings"],
            callbacks=[stats,hexdump,download,strings],
            context="mode"
            )
            
    def form(self,query,result):
        result.defaults = query
        result.case_selector()
        result.meta_selector(message='FS Image',case=query['case'],property='fsimage')
        result.textfield('Inode','inode')
        return result

class Timeline(Reports.report):
    """ View file MAC times in a searchable table """
    parameters = {'fsimage':'fsimage'}
    name = "View File Timeline"
    description = "Browse file creation, modification, and access times"

    def form(self, query, result):
        try:
            result.case_selector()
            if query['case']!=config.FLAGDB:
                result.meta_selector(message='FS Image',case=query['case'],property='fsimage')
        except KeyError:
            return result

    def analyse(self, query):
        dbh = self.DBO(query['case'])
        tablename = dbh.MakeSQLSafe(query['fsimage'])
        temp_table = dbh.get_temp()
        dbh.execute("create temporary table %s select i.inode,f.status,mtime as `time`,1 as `m`,0 as `a`,0 as `c`,0 as `d`,concat(path,name) as `name` from inode_%s as i left join file_%s as f on i.inode=f.inode" %
                    (temp_table, tablename, tablename));
        dbh.execute("insert into %s select i.inode,f.status,atime,0,1,0,0,concat(path,name) from inode_%s as i left join file_%s as f on i.inode=f.inode" % (temp_table, tablename, tablename))
        dbh.execute("insert into %s select i.inode,f.status,ctime,0,0,1,0,concat(path,name) from inode_%s as i left join file_%s as f on i.inode=f.inode" % (temp_table, tablename, tablename))
        dbh.execute("insert into %s select i.inode,f.status,dtime,0,0,0,1,concat(path,name) from inode_%s as i left join file_%s as f on i.inode=f.inode" % (temp_table, tablename, tablename))
        dbh.execute("create table if not exists mac_%s select inode,status,time,sum(m) as `m`,sum(a) as `a`,sum(c) as `c`,sum(d) as `d`,name from %s where time>0 group by time,name order by time,name" %
                    (tablename, temp_table))
        dbh.execute("alter table  mac_%s add key(inode)" % tablename)
        
        
    def progress(self, query, result):
        result.heading("Building Timeline")
    
    def display(self, query, result):
        dbh = self.DBO(query['case'])
        tablename = dbh.MakeSQLSafe(query['fsimage'])
        
        def DeletedIcon(value):
            """ Callback for deleted items """
            tmp=self.ui()
            if value=='alloc':
                tmp.icon("yes.png")
            elif value=='deleted':
                tmp.icon("no.png")
            else:
                tmp.icon("question.png")
            return tmp

        result.table(
            columns=('from_unixtime(time)','inode','status',
                     "if(m,'m',' ')","if(a,'a',' ')","if(c,'c',' ')","if(d,'d',' ')",'name'),
            names=('Timestamp', 'Inode','Del','m','a','c','d','Filename'),
            callbacks={'Del':DeletedIcon},
            table=('mac_%s' % tablename),
            case=query['case'],
#            links=[ None, None, None, None, None, None, None, FlagFramework.query_type((),case=query['case'],family=query['family'],fsimage=query['fsimage'],report='ViewFile',__target__='filename')]
            links=[ None, FlagFramework.query_type((),case=query['case'],family=query['family'],fsimage=query['fsimage'],report='ViewFile',__target__='inode')]
            )

    def reset(self, query):
        dbh = self.DBO(query['case'])
        tablename = dbh.MakeSQLSafe(query['fsimage'])
        dbh = self.DBO(query['case'])
        dbh.execute("drop table mac_%s" % tablename)


import md5

class HashComparison(BrowseFS):
    """ Compares MD5 hash against the NSRL database to classify files """
    parameters = {'fsimage':'fsimage'}
    name = "MD5 Hash comparison"
    description="This report will give a table for describing what the type of file this is based on the MD5 hash matches"
    progress_dict = {}

    def progress(self,query,result):
        dbh=self.DBO(query['case'])
        tablename = dbh.MakeSQLSafe(query['fsimage'])

        result.heading("Currently calculating NSRL Hash comparison")
        dbh.execute("select count(*) as `count` from file_%s where mode='r/r'",tablename)
        row=dbh.fetch()
        try:
            result.text("Just did %s out of %s files" % (self.progress_dict[tablename],row['count']))
        except KeyError:
            pass

    def display(self,query,result):
        result.heading("MD5 Hash comparisons for %s" % query['fsimage'])
        dbh=self.DBO(query['case'])
        tablename = dbh.MakeSQLSafe(query['fsimage'])

        def RenderNSRL(value):
            tmp=self.ui(result)
            if value>0:
                tmp.icon("yes.png")
            else:
                tmp.icon("no.png")

            return tmp

        result.table(
            columns=('Inode','Filename', '`File Type`', '`NSRL Product`','NSRL_filename', '`MD5`'),
            names=('Inode','Filename','File Type','NSRL Product','NSRLFilename','MD5'),
            table='hash_%s ' % (tablename),
            case=query['case'],
            links=[ FlagFramework.query_type((),case=query['case'],family=query['family'],fsimage=query['fsimage'],report='ViewFile',__target__='inode')]
            )

    def analyse(self,query):
        dbh = self.DBO(query['case'])
        tempname=dbh.get_temp()
        tablename=dbh.MakeSQLSafe(query['fsimage'])
        dbh.execute("""CREATE TABLE `%s` (
        `inode` varchar( 20 ) NOT NULL default '',
        `md5` varchar( 35 ) NOT NULL default '',
        `binary_md5` varchar( 16 ) binary NOT NULL default '',
        `filetype` tinytext NOT NULL ,
        `NSRL_productcode` int(11) not NULL default '0',
        `NSRL_filename` varchar(60) not NULL default ''
        )""" , tempname)

        iofd = IO.open(query['case'],query['fsimage'])
        ddfs = Sleuthkit.DDFS(query['case'],query['fsimage'],iofd)
        dbh_flag = self.DBO(None)
        dbh2=self.DBO(query['case'])

        ## We use this method to communicate with our other thread:
        self.progress_dict[tablename]=0
        
        magic = FlagFramework.Magic()
        dbh.execute("select path,name,inode from file_%s where mode='r/r'",tablename)        
        for row in dbh:
            self.progress_dict[tablename]+=1
            print "openning file %s%s"  % (row['path'],row['name'])
            try:
                fd = ddfs.open("%s%s" % (row['path'],row['name']))
                ## If the file is too fragmented, we dont take the hash for it, since it might take too long. This is mostly a problem with NTFS which suffers serious fragmentation issues, where files can be heavily fragmented leading to extremely long load times.
                if len(fd.blocks)>10000:
                    print "skipping, too many rows"
                    continue
            
                ## This is required in case we get a huge file:
                f=fd.read(1000000)
                sig=magic.buffer(f)
                m = md5.new(f)
                while 1:
                    f=fd.read(1000000)
                    if not f: break
                    m.update(f)

                fd.close()
                dbh_flag.execute("select filename,productcode from NSRL_hashes where md5=%r",m.digest())
                nsrl=dbh_flag.fetch()
                if not nsrl: nsrl={}          
                dbh2.execute("insert into %s set inode=%r,md5=%r,binary_md5=%r,filetype=%r,NSRL_productcode=%r,NSRL_filename=%r",(tempname,row['inode'],m.hexdigest(),m.digest(),sig,nsrl.get('productcode',''),nsrl.get('filename','')))

            except IOError,e:
                print "IOError on %s" % e
                continue

        ## Create the final hash table
        dbh.execute("create table `hash_%s` select a.inode as `Inode`,concat(path,b.name) as `Filename`,filetype as `File Type`,if(c.Code=0,'Unknown',c.Name) as `NSRL Product`,c.Code as `NSRL Code`,a.NSRL_filename,md5 as `MD5` from %s as a,%s.NSRL_products as c left join file_%s as b on a.inode=b.inode   where  a.NSRL_productcode=c.Code group by Inode,`NSRL Code`,MD5",(tablename,tempname,config.FLAGDB,tablename))
        dbh.execute("alter table `hash_%s` add index(`Inode`)",tablename)

    def reset(self,query):
        dbh = self.DBO(query['case'])
        tablename=dbh.MakeSQLSafe(query['fsimage'])
        dbh.execute("drop table `hash_%s`",tablename)
        
class BrowseRegistry(Reports.report):
    """ Browse a Windows Registry file """
    description="Browse a windows registry hive file (found in c:\winnt\system32\config\) "
    name = "Browse Registry Hive"
    parameters = {'regfile':'filename'}

    def display(self,query,result):
        result.heading("Registry Hive in file %r" % query['regfile'])
        dbh = self.DBO(query['case'])
        tablename = dbh.MakeSQLSafe(query['regfile'])
        new_q=query.clone()
            
        #Make a tree call back:
        def treecb(branch):
            """ This call back will render the branch within the registry file. """
            path ='/'.join(branch)
            dbh = self.DBO(query['case'])

            ##Show the directory entries:
            dbh.execute("select basename from regi_%s where dirname=%r and length(basename)>1",(tablename,path))
            for row in dbh:
                tmp=self.ui()
                tmp.link(row['basename'],new_q,mode='table',where_Path="%s/%s" %(path,row['basename']))
                yield(([row['basename'],tmp,'branch']))
                
            ## Show the Registry keys    
            #dbh.execute("select type as `Type`,reg_key as `Key`,if(length(value)<50,value,concat(left(value,50),' .... ')) as `Value` from reg_%s  where path=%r",(tablename,path))            
            #for row in dbh:
            #    tmp=self.ui()
            #    l=self.ui()
            #    l.link("%s " % row['Key'],query,mode="display",key=row['Key'],path=path)
            #    tmp.text(l)
            #    tmp.text(" (%s): "  % row['Type'],color='black')
            #    tmp.text(row['Value'],color='red',sanitize='full')
            #    tmp.text(finish=1,color='black')
            #    yield(([row['Key'],tmp,'leaf']))

        ## End Tree Callback
                
        try:
            if query['mode']=='table':
                del new_q['mode']
                for i in new_q.keys():
                    if i.startswith('where_'):
                        del new_q[i]
                        
                left=self.ui(result)
                left.link("View Tree",new_q)
                result.row(left)
                result.table(
                    columns=['path','type','reg_key','size','value'],
                    names=['Path','Type','Key','Size','Value'],
                    links=[ result.make_link(new_q,'open_tree',mark='target') ],
                    table='reg_%s'%tablename,
                    case=query['case'],
                    )
                
            elif query['mode']=='display':
                del new_q['mode']
                key = query['key']
                path=query['path']
                del new_q['key']
                del new_q['path']
                left=self.ui(result)
                left.link("View Tree",new_q)
                result.row(left)
                result.end_table()
                result.para("Key %s/%s:" % (path,key))

                def hexdump(query):
                    """ Show the hexdump for the key """
                    out = self.ui()
                    dbh.execute("select value from reg_%s where path=%r and reg_key=%r",(tablename,path,key))
                    row=dbh.fetch()
                    if row:
                        FlagFramework.HexDump(row['value'],out).dump()
                    return out
                        
                def strings(query):
                    """ Draw the strings in the key """
                    out = self.ui()
                    out.para("not implimented yet")
                    return out

                def stats(query):
                    """ display stats on a key """
                    out = self.ui()
                    out.para("not implimented yet")
                    return out
                    
                result.notebook(
                    names=["HexDump","Strings","Statistics"],
                    callbacks=[hexdump,strings,stats],
                    context="display_mode"
                    )


        except KeyError:
            ## Display tree output
            del new_q['mode']
            del new_q['open_tree']
            
            left=self.ui(result)
            left.link("View Table",new_q,mode='table')
            result.row(left)

            # display paths in tree
            tree = self.ui(result)
            tree.tree(cb=treecb,branch=[''])

            if not query.has_key('open_tree'):
                query['open_tree'] = '/';

            # now display keys in table
            new_q['mode'] = 'display'
            new_q['path'] = query['open_tree']
            
            table = self.ui(result)
            table.table(
                columns=['reg_key','type','size',"if(length(value)<50,value,concat(left(value,50),' .... '))"],
                names=('Key','Type','Size','Value'),
                table='reg_%s' % tablename,
                where="path=%r" % query['open_tree'],
                case=query['case'],
                links=[ result.make_link(new_q, 'key') ]
                )
            
        ## Now return the keys in the current branch:
            result.row(tree, table, valign="top")
    
    def form(self,query,result):
        tmp = self.ui(result)
        tmp.filebox(target='regfile')
        result.case_selector(message="Select Case to insert into:")
        result.row("Select Registry File to load:",tmp)

    def analyse(self,query):
        dbh = self.DBO(query['case'])
        tablename = dbh.MakeSQLSafe(query['regfile'])
                
        dbh.execute('create table if not exists reg_%s (`path` CHAR(250), `size` SMALLINT, `type` CHAR(12),`reg_key` VARCHAR(200),`value` text)',tablename)
        
        p_mysql = os.popen("/usr/bin/mysql -f -u %r -p%r -D%r"%(config.USER,config.PASSWD,query['case']),'w')
        p_reg_reader= os.popen("%s/regtool -f %s -t reg_%s" % (config.FLAG_BIN,query['regfile'],tablename))
        
        while 1:
            data= p_reg_reader.read(1000)
            if not data: break
            p_mysql.write(data)

        p_reg_reader.close()
        p_mysql.close()

        dbh.execute("alter table reg_%s add index(path(100))",tablename)
        ## Now create the directory indexes to speed up tree navigation:
        dbh.execute("create table regi_%s (`dirname` TEXT NOT NULL ,`basename` TEXT NOT NULL,KEY `dirname` (`dirname`(100)))",tablename)
        dirtable = {}
        dbh.execute("select path from reg_%s",tablename)
        for row in dbh:
            array=row['path'].split("/")
            while len(array)>1:
                new_dirname="/".join(array[:-1])
#                if not new_dirname: new_dirname='/'
                new_basename=array.pop()
                try:
                    ## See if the value is already in the dictionary
                    dirtable[new_dirname].index(new_basename)
                except ValueError:
                    dirtable[new_dirname].append(new_basename)
                except KeyError:
                    dirtable[new_dirname]=[new_basename]

        for k,v in dirtable.items():
            for name in v:
                dbh.execute("insert into regi_%s set dirname=%r,basename=%r",(tablename,k,name))


    def reset(self,query):
        dbh = self.DBO(query['case'])
        tablename = dbh.MakeSQLSafe(query['regfile'])
        
        dbh.execute('drop table if exists reg_%s',tablename)
        dbh.execute('drop table if exists regi_%s',tablename)

    def progress(self, query, result):
        result.heading("Loading Registry")
