# ******************************************************
# Copyright 2004: Commonwealth of Australia.
#
# Developed by the Computer Network Vulnerability Team,
# Information Security Group.
# Department of Defence.
#
# David Collett <daveco@users.sourceforge.net>
# Michael Cohen <scudette@users.sourceforge.net>
#
# ******************************************************
#  Version: FLAG 0.4 (12-02-2004)
# ******************************************************
#
# * This program is free software; you can redistribute it and/or
# * modify it under the terms of the GNU General Public License
# * as published by the Free Software Foundation; either version 2
# * of the License, or (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ******************************************************

""" Module for analysing Log files """
import pyflag.Reports as Reports
import pyflag.FlagFramework as FlagFramework
import pyflag.LogFile as LogFile
import pyflag.conf
config=pyflag.conf.ConfObject()
import re

description = "Log Analysis"
order = 35

class ListLogFile(Reports.report):
    """ Lists the content of the log file using the table UI object """
    parameters = {"logtable":"casetable"}
    name="List log file contents"
    description="This report simply lists the log entries in a searchable/groupable table"

    def form(self,query,result):
        result.case_selector()
        result.meta_selector(query['case'],'Select Log Table','logtable')

    def display(self,query,result):
        #result.defaults = query
        #Find out the proper names of the field
        dbh = self.DBO(query['case'])

        #Find the names of all the columns in table:
        dbh.execute("select * from %s limit 1",query['logtable'])
            
        columns =[]
        for d in dbh.cursor.description:
            columns.append(d[0])

        result.table(columns=columns,names=columns,links=[], table=query['logtable'], case=query['case'])

class CreateLogPreset(Reports.report):
    """ Creates a new type of log file in the database, so that they can be loaded using the Load Log File report """
    parameters = {"log_preset":"sqlsafe", "datafile":"filename",
                  "field":"unique", "final":"alphanum", "preview":"alphanum"}
    name="Create Log Preset"
    description="Create new preset log type"
    order=40

    def display(self,query,result):
        result.heading("New log file preset %s created" % query['log_preset'])
        result.link("Load a log file", FlagFramework.query_type((),case=query['case'],family='LoadData',report='LoadPresetLog',log_preset=query['log_preset']))
        return result

    def form(self,query,result):
        try:
            result.start_table()
            result.ruler()
            #Check if the user has supplied a name for the table: (if they did not we throw an exception and return the form so far - the overall effect is that the form will grow as the user enters more data.
            
            tmp = self.ui()
            tmp.heading("Step 1:")
            result.row(tmp,"Select a sample log file for the previewer")
            result.ruler()
            #Create a new ui based on our current one
            tmp = self.ui(default = result)
            tmp.filebox()
            result.row("Enter name of log file:",tmp)

            result.ruler()
            tmp = self.ui()
            tmp.heading("Step 2:")
            result.row(tmp,"Select a field Separator")
            result.ruler()

            #If there is no delimiter method - we choose simple
            try:
                query['delmethod']
            except KeyError:
                query['delmethod'] = 'simple'

            result.radio("Field Separator Type:","delmethod",('simple','advanced'),onclick="this.form.submit();")

            if query['delmethod'] == 'simple':
                result.selector("Simple Field Separator:",'delimiter',"select delimiter,description from simple_delimiters",(),case=None)
            elif query['delmethod'] == 'advanced':
                result.textfield("Enter Regular Expression:","delimiter")

            if not query.has_key('delimiter'):
                query['delimiter'] = ' '
            # get a log loader
            log = LogFile.loader(query['datafile'],query['delmethod'],query['delimiter'])

            # preview raw file
            result.row("Unprocessed text from file",colspan=5)
            log.preview(result,mode='raw')
           
            #Step 3:
            result.ruler()
            tmp = self.ui(result)
            tmp.heading("Step 3:")
            result.row(tmp,"Select pre-filter(s) to use on the data")
            result.ruler()
            LogFile.pre_selector(result)

            #Now show the user what the pre-filters do on the data:
            log.set_prefilters(query)
            log.preview(result,mode='filtered')
   
            result.end_table()
            result.start_table()
                                 
            #Step 4:
            result.ruler()
            tmp = self.ui(result)
            tmp.heading("Step 4:")
            result.row(tmp,"Assign field names and Types to each field",align='left')
            result.ruler()

            #result.end_table()
            #result.start_table()
            
            #Show the split up lines:
            tmp_table = self.ui(result)
            tmp_table.start_table(border='1')
            log.set_fields(query)
            log.preview(tmp_table, ignore=False, headings=False)
            
            # allow the user to give names and datatypes for each field
            # we use 'field%u' because order is important, an array from HTTP wont do.
            field = []
            type = []
            index = []
            for i in range(log.num_fields(ignore=False)):
                # set the default fieldname to ignore
                try:
                    query["field%u"% i]
                except KeyError:
                    query["field%u"% i] = 'ignore'
                    
                field_ui = self.ui(tmp_table)
                type_ui = self.ui(tmp_table)
                index_ui = self.ui(tmp_table)
                field_ui.textfield('','field%u'%i)
                LogFile.type_selector(type_ui,'type%u'%i)
                index_ui.checkbox('Add Index?','index%u'%i,'yes')
                field.append(field_ui)
                type.append(type_ui)
                index.append(index_ui)

            tmp_table.row(*field)
            tmp_table.row(*type)
            tmp_table.row(*index)
            tmp_table.end_table()
            result.row(tmp_table,bgcolor='lightgray',colspan=5)
            result.hidden('field','ok')
            
            result.end_table()
            result.start_table()

            result.checkbox('Click here and submit to see final preview','preview','ok')
            query['preview']

            # set types
            log.set_types(query)
            
            # try to load and display as a final test
            dbh = self.DBO(query['case'])
            temp_table = dbh.get_temp()
            
            log.load(query['case'],query['datafile'],temp_table, rows=3)
            dbh.execute("select * from %s limit 1",temp_table)
            columns =[]
            for d in dbh.cursor.description:
                columns.append(d[0])

            result.ruler()
            tmp_final = self.ui(result)
            tmp_final.table(columns=columns,names=columns,links=[], table=temp_table, case=query['case'], simple=True)
            result.row(tmp_final,bgcolor='lightgray',colspan=5)

            result.ruler()
            tmp = self.ui(result)
            tmp.heading("Step 5:")
            result.row(tmp,"Give the preset a name",align='left')
            result.ruler()
            result.textfield("name for preset:",'log_preset')
            result.ruler()
            result.checkbox('Click here when finished','final','ok')
            result.end_table()

            return result

        #If any of the required paramters are not there, we pass the form back to the user to complete.
        except KeyError,e:
            return result

    def analyse(self, query):
        """ store this log type in the database """
        if not query.has_key('delimiter'):
            query['delimiter'] = ' ';
            
        # try to make a loader object
        log = LogFile.loader(query['datafile'], query['delmethod'], query['delimiter'])
        log.set_prefilters(query)
        log.set_fields(query)
        log.set_types(query)
        log.set_indexes(query)

        # pickle it to the database
        LogFile.store_loader(log, query['log_preset'])
