# ******************************************************
# Copyright 2004: Commonwealth of Australia.
#
# Developed by the Computer Network Vulnerability Team,
# Information Security Group.
# Department of Defence.
#
# David Collett <daveco@users.sourceforge.net>
#
# ******************************************************
#  Version: FLAG 0.4 (12-02-2004)
# ******************************************************
#
# * This program is free software; you can redistribute it and/or
# * modify it under the terms of the GNU General Public License
# * as published by the Free Software Foundation; either version 2
# * of the License, or (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ******************************************************

""" Module for handling Log Files """
import pyflag.Reports as Reports
import pyflag.FlagFramework as FlagFramework
import pyflag.DB as DB
import pyflag.conf
config=pyflag.conf.ConfObject()
import re

class prefilter:
    """ This class defines all the prefilters that are appropriate for importing log files.
    
    Prefilters are of the prototype:
    
    string prefilter(string)
    
    Method names must start with \"PF\", this will allow the class to pick its methods by itself. The use of a short, 1 line docstring is mandatory, since this is how the class will describe the prefilter to the user.
    
    @ivar  filters: A dict mapping the filter methods to their docstring descriptions.
    @ivar  res: A dict managing lists of (compiled RE's,target strings). This way REs only need to be compiled once.
    """
    
    filters = {}
    res = {}
    
    def __init__(self):
        for a in dir(self):
            if a.startswith('PF'):
                self.filters[a] = (prefilter.__dict__[a],prefilter.__dict__[a].__doc__)

    def transform(self, transform_list,string):
        """ Transforms a string according to a list of transformation.

        @arg transform_list: List of transformations obtained from prepare()
        @arg string: string to transform
        @return: A transformed string """
        for re_expression,target in transform_list:
            string = re_expression.sub(target,string)

        return string


    def prepare(self,re_strings,list):
        """ prepares a string and pushes it onto a list.

        This function can handle perl-like s/.../../ operations in a naive manner. There may be multiple lines of these expressions, in which case- the res will be compiled one line at the time and inserted into the list. Note that the s parsing is simplistic and assumes that the delimiter is after the first letter of the expression which must be an 's'. We assume that the delimiter can not be escaped. If you need to escape the delimiter - use another delimiter.

        @arg re_strings: A possibly multi-lined string of the style: s/.../.../
        @arg list: The list we push the (re,sub_string) onto.
        """
        for i in re_strings.splitlines():
            i = i.strip()
            if not i: continue
            if i[0] != 's': raise REError, "Regular expressions must start with s"
            delimiter = i[1]
            tmp = i.split(delimiter)
            try:
                if tmp[3]:
                    tmp[1] = "(?"+tmp[3]+")" + tmp[1]
            except KeyError:
                pass
            
            list.append((re.compile(tmp[1]),tmp[2]))

    def PFDateFormatChange(self,string):
        """ DD/MM/YYYY->YYYY/MM/DD """
        if not self.res.has_key('PFDateFormatChange'):
            tmp = []
            self.prepare(r" s#(\d\d)\/([^\/]+)\/(\d\d\d\d)#\3/\2/\1# ",tmp)
            self.res['PFDateFormatChange'] = tmp
            
        transform_list = self.res['PFDateFormatChange']
        return self.transform(transform_list,string)

    def PFDateFormatChange2(self,string):
        """ YYYY-MM-DD HH:MM:SS->YYYY/MM/DD:HH:MM:SS """
        if not self.res.has_key('PFDateFormatChange2'):
            tmp=[]
            self.prepare(r"s|(\d\d\d\d)-(\d\d)-(\d\d) (\d\d:\d\d:\d\d)|\1/\2/\3:\4|" , tmp)
            self.res['PFDateFormatChange2'] = tmp

        transform_list = self.res['PFDateFormatChange2']
        return self.transform(transform_list,string)

    def PFDateConvert(self,string):
        """ Month name to numbers """
        if not self.res.has_key('PFDateConvert'):
            tmp = []
            self.prepare(""" s/Jan(uary)?/1/i 
            s/Feb(uary)?/2/i
            s/Mar(ch)?/3/i
            s/Apr(il)?/4/i
            s/May/5/i
            s/Jun(e)?/6/i
            s/Jul(y)?/7/i
            s/Aug(ust)?/8/i
            s/Sep(tember)?/9/i
            s/Oct(ober)?/10/i
            s/Nov(ember)?/11/i
            s/Dec(ember)?/12/i
            """,tmp)
            self.res['PFDateConvert'] = tmp

        transform_list = self.res['PFDateConvert']
        return self.transform(transform_list,string)

    def PFRemoveChars(self,string):
        """ Remove []\"' chars """
        if not self.res.has_key('PFRemoveChars'):
             tmp = []
             self.prepare(" s/[\[\]\'\"]/ / ",tmp)
             self.res['PFRemoveChars'] = tmp

        return self.transform(self.res['PFRemoveChars'],string)


class loader:
    """ This class defines an object used to load log file, and perform tasks like displaying log previews
    
    @ivar  delmethod: Delimiter type string, either 'simple' or 'advanced' (for regex)
    @ivar  delimiter: Either a single character, or a string containing a regular expression
    @ivar  prefilters: List of prefilters to apply to data
    @ivar  fields:  List of field names, a field name of 'ignore' means that field will be ignored
    @ivar  types:  List of field data types
    @ivar  indexes: Index of colums to create indexes on in the database
    """

    def __init__(self, datafile, delmethod=None, delimiter=None, prefilters=None, fields=None, types=None, indexes=None):
        self.delmethod = 'simple'
        self.delimiter = ' '
        self.prefilters = []
        self.fields = []
        self.types = []
        self.indexes = []
        self.sample = None
        self.datafile = ''
        
        if delmethod:
            self.delmethod = delmethod
        if delimiter:
            self.delimiter = delimiter
        if prefilters:
            self.prefilters = prefilters
        if fields:
            self.fields = fields
        if datafile:
            self.datafile = datafile
        if types:
            self.types = types
        if indexes:
            self.indexes = indexes

    def set_prefilters(self, query):
        """ set the prefilters from the query """
        self.prefilters = query.getarray('prefilter')

    def set_fields(self, query):
        """ set the field names from the query, order is important """
        for i in range(self.num_fields(ignore=False)):
            try:
                self.fields.append(query['field%u'%i])
            except KeyError:
                self.fields.append('ignore')

    def set_types(self, query):
        """ set the field types from the query, order is important """
        for i in range(self.num_fields(ignore=False)):
            try:
                self.types.append(query['type%u'%i])
            except KeyError:
                self.types.append('varchar(250)')

    def set_indexes(self, query):
        """ which fields require indexes """
        for i in range(self.num_fields(ignore=False)):
            if query.has_key('index%u'%i):
                self.indexes.append(True)
            else:
                self.indexes.append(False)

    def get_fields(self, ignore=True):
        if ignore:
            return [ f for f in self.fields if f != 'ignore' ]
        else:
            return self.fields
                
    def field(self, index):
        try:
            name = self.fields[index]
        except IndexError:
            name = 'ignore'
        return name

    def get_sample(self):
        """ read 3 lines (minus comments) into sample string"""
        fd = open(self.datafile,'r')
        count = 3
        sample = ''
        for line in fd:
            line = line.strip()
            #Dont count lines starting with a comment or empty lines
            if not line or line.startswith('#'): continue
            
            sample+=line+"\n"
            count -=1
            if count <=0:
                break
        self.sample = sample

    def num_fields(self, ignore=True):
        if self.sample==None:
            self.get_sample()
        return len(self.split_line(self.sample.splitlines()[0], ignore))

    def split_line(self, line, ignore=True):
        """ split the given line based on delimiters, return array of fields """
        if self.delmethod == 'simple':
            line_array = line.split(self.delimiter)
            if ignore:
                line_array = [ line_array[i] for i in range(len(line_array)) if self.field(i) != 'ignore' ]
        elif self.delmethod == 'advanced':
            regex = re.compile(self.delimiter)
            m = regex.match(line)
            line_array = m.groups()
            if ignore:
                line_array = [ line_array[i] for i in range(len(line_array)) if self.field(i) != 'ignore' ]
        return line_array
       
    def preview(self, result, mode='full', ignore=True, headings=True):
        """ draw a preview of the logfile in datafile to result"""

        # read 3 lines from the file
        if self.sample==None:
            self.get_sample()

        if mode == 'raw':
            result.row(self.sample,bgcolor='lightgray',colspan=5)
            return

        # apply prefilters
        self.sample = self.filter(self.sample)
        if mode == 'filtered':
            result.row(self.sample,bgcolor='lightgray',colspan=5)
            return

        # split into fields
        if headings:
            result.row(*self.get_fields(ignore))
        for line in self.sample.splitlines():
            result.row(*self.split_line(line, ignore))

    def filter(self,string):
        """ Apply the transformations specified in query's prefilter array on the string.

        @arg string: String data to operate on
        @arg query: A query_type object which may contain some prefilter parameters
        @return: A transformed string """
        p = prefilter()
        for i in self.prefilters:
            #Call the relevant methods on the prefilter object:
            string = p.filters[i][0](p,string)

        return string
        

    def load(self, case, datafile, tablename, rows=None):
        """ Load log file data into table 'tablename', create it first if it doesn't exist"""
        dbh = DB.DBO(case)
        
        # create the table
        cols = []
        for i in range(len(self.fields)):
            if self.fields[i] != 'ignore':
                cols.append("%s %s" % (self.fields[i], self.types[i]))
        
        dbh.execute("CREATE TABLE IF NOT EXISTS %s (id int auto_increment,%s,key(id))" % (tablename,",".join(cols)))
        # prepare the insert string
        insert_str = "INSERT INTO "+tablename+" values(Null"+',%r'*self.num_fields()+")"
        
        # load the data
        count = 0
        fd = open(datafile,'r')
        for line in fd:
            if rows and count >= rows:
                break
            
            line = line.strip()
            #Dont count lines starting with a comment or empty lines
            if not line or line.startswith('#'): continue
            
            #apply filters
            line = self.filter(line)
            
            #split into fields and insert
            dbh.execute(insert_str,self.split_line(line))
            count += 1

        # now add indexes
        for i in range(len(self.indexes)):
            if self.indexes[i]:
                dbh.execute("ALTER TABLE %s add key(%s)"%(tablename, self.fields[i]))

    def __str__(self):
        str = ''
        str += "delmethod: %s\n" % self.delmethod
        str += "delimiter: %s\n" % self.delimiter
        str += "prefilers: %s\n" % ','.join(self.prefilters)
        str += "fields: %s\n" % ','.join(self.fields)
        str += "types: %s\n" % ','.join(self.types)
        str += "indexes: %s\n" % ','.join(self.indexes)
        return str

import pickle

def get_loader(name):
    """ lookup and unpickle log loader from the database, return loader object """
    dbh = DB.DBO(config.FLAGDB)
    return loader(*pickle.loads(dbh.get_meta('log_preset_%s' % name)))
    
def store_loader(loader, name):
    """ pickle and save given loader into the database """
    preset = (loader.datafile, loader.delmethod, loader.delimiter, tuple(loader.prefilters),
              tuple(loader.fields), tuple(loader.types), tuple(loader.indexes))
    dbh = DB.DBO(config.FLAGDB)
    dbh.execute('INSERT INTO meta values ("log_preset", %r)' % name)
    dbh.execute('INSERT INTO meta values ("log_preset_%s", %r)' % (name, pickle.dumps(preset)))

def pre_selector(result):
    f = prefilter().filters
    x = []
    y = []
    for i in f.keys():
        x.append(i)
        y.append(f[i][1])
        
    result.const_selector("pre-filter(s) to use on the data:",'prefilter',x,y,size=4,multiple="multiple")

types = ('varchar(250)','int','datetime','text')

def type_selector(result, name):
    result.const_selector('',name, types, types)
