# ******************************************************
# Copyright 2004: Commonwealth of Australia.
#
# Developed by the Computer Network Vulnerability Team,
# Information Security Group.
# Department of Defence.
#
# Michael Cohen <scudette@users.sourceforge.net>
# David Collett <daveco@users.sourceforge.net>
#
# ******************************************************
#  Version: FLAG 0.4 (12-02-2004)
# ******************************************************
#
# * This program is free software; you can redistribute it and/or
# * modify it under the terms of the GNU General Public License
# * as published by the Free Software Foundation; either version 2
# * of the License, or (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ******************************************************

""" Module used to interface to the flag version of sleuthkit
"""
import os,os.path
import pyflag.conf
config=pyflag.conf.ConfObject()

import pyflag.DB as DB
import pyflag.IO as IO
import time
import math
import bisect
import os.path

def load_sleuth(case,fstype,table,iosource):
    """ Loads the filesystem data into the case """

    # get the IO object to load, we dont really need to do this,
    # but its the easiest way to get the option string and helps to
    # ensure the next step will actually work.
    fd = IO.open(case,iosource)
    opts = ""
    for i in range(len(fd.parameters)-1):
        for j in range(len(fd.options[i+1])):
            opts += "%s=%s " % (fd.parameters[i+1][3:], fd.options[i+1][j])

    pipe = os.popen("%s -t %s -d create blah" %(config.SLEUTHKIT,table),'r')
    sql_schema = pipe.read()
    pipe.close()

    p_sleuth = os.popen("%s -t %s -f %s -i %s %s"%(config.SLEUTHKIT,table,fstype,fd.options[0][0],opts),'r')
    p_mysql = os.popen("/usr/bin/mysql -f -u %r -p%r -D%r"%(config.USER,config.PASSWD,case),'w')

    p_mysql.write(sql_schema)
    while 1:
        data= p_sleuth.read(10000)
        if not data: break
        p_mysql.write(data)

    p_sleuth.close()
    p_mysql.close()

def del_sleuth(case, table):
    """ Drops the tables for the given iosource identifier """
    pipe = os.popen("%s -t %s -d drop" %(config.SLEUTHKIT,table),'r')
    drop_cmd = pipe.read()
    pipe.close()
    p_mysql = os.popen("/usr/bin/mysql -f -u %r -p%r -D%r"%(config.USER,config.PASSWD,case),'w')
    p_mysql.write(drop_cmd)
    p_mysql.close()

class DDFS:
    """ Class for accessing filesystems in a DD file using data in the database """
    def __init__(self, case, table, fd):
        """ Initialise the DDFS object """
        self.fd = fd
        self.table = table
        self.case = case
        self.dbh = DB.DBO(case)

    def longls(self,path='/'):
        """ list directory content longly """
        self.dbh.execute("select mode,inode,name from file_%s where path=%r", (self.table, path))
        return [ dent for dent in self.dbh ]
    
    def ls(self, path="/", dirs=None):
        """list directory contents"""
        if(dirs == 1):
            self.dbh.execute("select name from file_%s where path=%r and mode='d/d'", (self.table, path))
        elif(dirs == 0):
            self.dbh.execute("select name from file_%s where path=%r and mode='r/r'", (self.table, path))
        else:
            self.dbh.execute("select name from file_%s where path=%r", (self.table, path))
        return [ dent['name'] for dent in self.dbh ]

    def dent_walk(self, path='/'):
        """return directory entries in the given path, one at a time, ONLY USE FILE TABLE"""
        self.dbh.execute("select name, mode, status from file_%s where path=%r order by name" % (self.table, path))
        for i in self.dbh:
            yield(i)
    
    def lookup(self, path=None,inode=None):
        """return the inode number for the given path"""
        if path:
            dir,name = os.path.split(path)
            if not name:
                dir,name = os.path.split(path[:-1])
            if dir == '/':
                dir = ''

            self.dbh.execute("select inode from file_%s where path=%r and (name=%r or name=concat(%r,'/'))", (self.table,dir+'/',name,name))
            res = self.dbh.fetch()
            if not res:
                return None
            return res["inode"]
        else:
            self.dbh.execute("select concat(path,name) as path from file_%s where inode=%r", (self.table,inode))
            res = self.dbh.fetch()
            if not res:
                return None
            return res["path"]
    
    def open(self, path=None, inode=None):
        if not inode:
            inode = self.lookup(path)
        if not inode:
            raise IOError('Inode not found for file')
        return DDFS_file(self.case, self.table, self.fd, inode)

    def istat(self, path=None, inode=None):
        if not inode:
            inode = self.lookup(path)
        if not inode:
            return None
        self.dbh.execute("select inode, status, uid, gid, from_unixtime(mtime) as `mtime`, from_unixtime(atime) as `atime`, from_unixtime(ctime) as `ctime`, from_unixtime(dtime) as `dtime`, mode, links, link, size from inode_%s where inode=%r",(self.table, inode))
        return self.dbh.fetch()

    def isdir(self,directory):
        """ Returns 1 if directory is a directory, 0 otherwise """
        if not directory.endswith('/'):
            directory+='/'
        self.dbh.execute("select mode from file_%s where path=%r",(self.table,directory))
        row=self.dbh.fetch()
        if row:
            return 1
        else:
            return 0
        
    def exists(self,path):
        """ Returns 1 if path exists, 0 otherwise """
        dir,file=os.path.split(path)
        self.dbh.execute("select mode from file_%s where path=%r and name=%r",(self.table,dir,file))
        row=self.dbh.fetch()
        if row:
            return 1
        else:
            return 0
        
class DDFS_file:
    """ Class for reading files within a loaded dd image, supports typical file-like operations, seek, tell, read """
    def __init__(self, case, table, fd, inode):
        self.dbh = DB.DBO(case)
        self.fd = fd
        self.table = table
        self.readptr = 0
        self.dbh.execute("select value from meta_%s where name='block_size'",self.table);
        self.block_size = int(self.dbh.fetch()["value"])
        # fetch inode data
        self.dbh.execute("select * from inode_%s where inode=%r", (self.table, inode))
        self.data = self.dbh.fetch()
        if not self.data:
            raise IOError("Could not locate inode %s"% inode)

        self.size = self.data['size']
        self.dbh.execute("select block,count,`index` from block_%s where inode=%r order by `index`", (self.table, inode))
        try:
            self.blocks = [ (row['block'],row['count'],row['index']) for row in self.dbh ]
        except KeyError:
            self.blocks = None
        self.index = [ d[2] for d in self.blocks ]
        
    def close(self):
        """ Fake close method. """
        pass

    def getval(property):
        try:
            return self.data[property]
        except KeyError:
            return None
        
    def seek(self, offset, rel=None):
        """ fake seeking routine, doesnt really seek, just updates the read pointer """
        if(rel):
            self.readptr += offset
        else:
            self.readptr = offset
        if(self.readptr > self.size):
            self.readptr = self.size

    def read(self, length=None):
        if (length == None) or ((length + self.readptr) > self.size):
            length = self.size - self.readptr

        if length == 0:
            return ''

        if not self.blocks:
            # now try to find blocks in the resident table
            self.dbh.execute("select data from resident_%s where inode=%r" % (self.table, self.data['inode']));
            row = self.dbh.fetch()
            if not row:
                raise IOError("Cant find any file data")
            data = row['data'][self.readptr:length+self.readptr]
	    self.readptr += length
	    return data

        fbuf=''
        while length>0:
        ## Where are we in the chunk?
            ddoffset,bytes_left = self.offset(self.readptr)
            
            self.fd.seek(ddoffset)
            if(bytes_left > length):
                fbuf += self.fd.read(length)
                self.readptr+=length
                return fbuf
            else:
                fbuf += self.fd.read(bytes_left)
                length-=bytes_left
                self.readptr+=bytes_left

        return fbuf

        
    def tell(self):
        """ returns the current read pointer"""
        return self.readptr

    def offset(self,offset):
        """ returns the offset into the current block group where the given offset is found"""
        ## The block in the file where the offset is found
        block = int(offset/self.block_size)

        ##Obtain the index of blocks array where the chunk is. This is the index at which self.index is 
        blocks_index=0
        try:
            while 1:
                if self.index[blocks_index]<=block<self.index[blocks_index+1]: break
                blocks_index+=1

        except IndexError:
            blocks_index=len(self.index)-1

        #If the end of the chunk found occurs before the block we seek, there is something wrong!!!
        if self.blocks[blocks_index][1]+self.blocks[blocks_index][2]<=block:
            raise IOError("Block table does not span seek block %s"%block,offset)

        ## Look the chunk up in the blocks array
        ddblock,count,index=self.blocks[blocks_index]

        ## The offset into the chunk in bytes
        chunk_offset = offset-index*self.block_size

        ## The dd offset in bytes
        ddoffset=ddblock*self.block_size+chunk_offset

        ## The number of bytes remaining in this chunk
        bytes_left = count*self.block_size-chunk_offset
        
        return ddoffset,bytes_left
